# Copyright 2015 Cisco Systems.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import collections

import mock

from neutron.tests import base

from networking_cisco.apps.saf.agent.vdp import ovs_vdp
from networking_cisco.apps.saf.agent.vdp import vdp_constants as vconstants
from networking_cisco.apps.saf.common import dfa_sys_lib as utils

try:
    OrderedDict = collections.OrderedDict
except AttributeError:
    import ordereddict
    OrderedDict = ordereddict.OrderedDict


class OvsVdpTest(base.BaseTestCase):
    """A test suite to exercise the OvsVdp Class.  """

    def setUp(self):
        '''Setup routine '''
        super(OvsVdpTest, self).setUp()
        self.root_helper = 'sudo'
        self.uplink = "eth2"
        self.integ_br = 'br-int1'
        self.ext_br = 'br-ethd1'

        self.port_name = "loc_veth"
        self.port_str = "loc_veth_eth2"
        self.rpc_client = mock.Mock()
        self.execute = mock.patch.object(
            utils, "execute", spec=utils.execute).start()
        self._test_ovs_vdp_init()
        phy_port_num = 5
        int_peer_port_num = 6
        self.port_uuid = '0000-1111-2222-3334'
        self.port_uuid_1 = '0000-1111-2222-3335'
        self.mac = '00:00:fa:11:22:34'
        self.net_uuid = '0000-aaaa-bbbb-cccc'
        self.segmentation_id = 10001
        self.lvid = 10
        self.exist_vdp_vlan = 3000
        self.oui = None
        self.ovs_vdp.phy_peer_port_num = phy_port_num
        self.ovs_vdp.int_peer_port_num = int_peer_port_num

    def _test_ovs_vdp_init(self):
        '''Test the init routine '''
        lldp_ovs_portnum = 14
        phy_port_num = 14
        get_br = mock.patch('networking_cisco.apps.saf.common.'
                            'dfa_sys_lib.get_bridges').start()
        is_patch = mock.patch('networking_cisco.apps.saf.common.'
                              'dfa_sys_lib.is_patch').start()
        get_peer = mock.patch('networking_cisco.apps.saf.common.'
                              'dfa_sys_lib.get_peer').start()
        self.ovs_br_add = mock.patch('networking_cisco.apps.saf.common.'
                                     'dfa_sys_lib.OVSBridge.add_flow').start()
        mock.patch('networking_cisco.apps.saf.common.'
                   'dfa_sys_lib.OVSBridge.add_port',
                   return_value=str(lldp_ovs_portnum)).start()
        mock.patch('networking_cisco.apps.saf.common.'
                   'dfa_sys_lib.OVSBridge.get_port_ofport',
                   return_value=str(phy_port_num)).start()
        port_name_list = mock.patch('networking_cisco.apps.saf.common.'
                                    'dfa_sys_lib.OVSBridge.'
                                    'get_port_name_list').start()
        dev_exist = mock.patch('neutron.agent.linux.ip_lib.'
                               'device_exists').start()
        add_veth = mock.patch('neutron.agent.linux.ip_lib.'
                              'IPWrapper.add_veth').start()
        mock.patch('networking_cisco.apps.saf.'
                   'agent.vdp.lldpad.LldpadDriver').start()
        mock.patch('networking_cisco.apps.saf.common.'
                   'utils.PeriodicTask').start()
        get_br.return_value = self.integ_br + ' ' + self.ext_br
        port_name_list.return_value = 'port_int1'
        is_patch.return_value = True
        get_peer.return_value = 'port_int1'
        dev_exist.return_value = False
        add_veth.return_value = mock.Mock(), mock.Mock()
        self.vdp_cb = mock.MagicMock()
        self.ovs_vdp = ovs_vdp.OVSNeutronVdp(self.uplink, self.integ_br,
                                             self.ext_br,
                                             self.root_helper, self.vdp_cb)
        self.ovs_br_add.reset_mock()

    def test_process_init(self):
        '''Wrapper for the init routine test '''
        pass

    def test_setup_lldpad_ports(self):
        """Test for setup lldpad ports. """
        lldp_ovs_portnum = 14
        phy_port_num = 14
        veth_str = vconstants.LLDPAD_LOC_VETH_PORT + self.uplink
        with mock.patch('networking_cisco.apps.saf.'
                        'agent.vdp.lldpad.LldpadDriver') as lldpad, \
            mock.patch('networking_cisco.apps.saf.common.dfa_sys_lib.'
                       'OVSBridge.delete_flows') as ovs_br_del:
            lldp_inst = lldpad.return_value
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_del, 'delete_flows')
            parent.attach_mock(self.ovs_br_add, 'add_flow')
            parent.attach_mock(lldp_inst.enable_evb, 'enable_evb')
            self.ovs_vdp.setup_lldpad_ports()
        expected_calls = [
            mock.call.delete_flows(dl_dst=vconstants.NCB_DMAC,
                                   dl_type=vconstants.LLDP_ETYPE),
            mock.call.delete_flows(dl_dst=vconstants.NCB_DMAC,
                                   dl_type=vconstants.VDP22_ETYPE),
            mock.call.add_flow(priority=vconstants.VDP_FLOW_PRIO,
                               in_port=str(lldp_ovs_portnum),
                               dl_dst=vconstants.NCB_DMAC,
                               dl_type=vconstants.LLDP_ETYPE,
                               actions="output:%s" % phy_port_num),
            mock.call.add_flow(priority=vconstants.VDP_FLOW_PRIO,
                               in_port=str(phy_port_num),
                               dl_dst=vconstants.NCB_DMAC,
                               dl_type=vconstants.LLDP_ETYPE,
                               actions="output:%s" % lldp_ovs_portnum),
            mock.call.add_flow(priority=vconstants.VDP_FLOW_PRIO,
                               in_port=str(lldp_ovs_portnum),
                               dl_dst=vconstants.NCB_DMAC,
                               dl_type=vconstants.VDP22_ETYPE,
                               actions="output:%s" % phy_port_num),
            mock.call.add_flow(priority=vconstants.VDP_FLOW_PRIO,
                               in_port=str(phy_port_num),
                               dl_dst=vconstants.NCB_DMAC,
                               dl_type=vconstants.VDP22_ETYPE,
                               actions="output:%s" % lldp_ovs_portnum),
            mock.call.enable_evb()]
        parent.assert_has_calls(expected_calls, any_order=False)
        lldpad.assert_called_with(veth_str, self.uplink, self.root_helper)

    def _test_vdp_port_event_new(self):
        '''Test the case for a new vnic port for a network '''
        port_uuid = '0000-1111-2222-3333'
        mac = '00:00:fa:11:22:33'
        net_uuid = '0000-aaaa-bbbb-cccc'
        segmentation_id = 10001
        status = 'up'
        oui = None
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.get_ofport_name',
                        return_value='test_port'), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.get_port_vlan_tag',
                       return_value=10), \
            mock.patch.object(self.ovs_vdp.lldpad_info, 'send_vdp_vnic_up',
                              return_value=[500, None]):
            parent = mock.MagicMock()
            parent.attach_mock(self.ovs_br_add, 'add_flow')
            self.ovs_vdp.send_vdp_port_event(port_uuid, mac, net_uuid,
                                             segmentation_id, status, oui)
        expected_calls = [
            mock.call.add_flow(priority=4,
                               in_port=self.ovs_vdp.phy_peer_port_num,
                               dl_vlan=10,
                               actions="mod_vlan_vid:%s,normal" % 500),
            mock.call.add_flow(priority=3,
                               in_port=self.ovs_vdp.int_peer_port_num,
                               dl_vlan=500,
                               actions="mod_vlan_vid:%s,normal" % 10)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def _test_vdp_port_event_exist(self):
        '''Test the case for a exist vnic port for a network '''
        port_uuid = '0000-1111-2222-3334'
        mac = '00:00:fa:11:22:34'
        net_uuid = '0000-aaaa-bbbb-cccc'
        segmentation_id = 10001
        status = 'up'
        oui = None
        ovs_cb_data = {'obj': self.ovs_vdp, 'mac': mac,
                       'port_uuid': port_uuid, 'net_uuid': net_uuid}
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.get_ofport_name',
                        return_value='test_port'), \
                mock.patch.object(self.ovs_vdp.lldpad_info,
                                  'send_vdp_vnic_up') as vnic_up:
            parent = mock.MagicMock()
            parent.attach_mock(vnic_up, 'send_vdp_vnic_up')
            self.ovs_vdp.send_vdp_port_event(port_uuid, mac, net_uuid,
                                             segmentation_id, status, oui)
        expected_calls = [
            mock.call.send_vdp_vnic_up(port_uuid=port_uuid,
                                       vsiid=port_uuid,
                                       gid=segmentation_id,
                                       mac=mac, vlan=500,
                                       oui=oui,
                                       vsw_cb_fn=self.ovs_vdp.vdp_vlan_change,
                                       vsw_cb_data=ovs_cb_data)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def _test_vdp_port_event_down(self):
        '''Test the case for a vnic port down for a network '''
        port_uuid = '0000-1111-2222-3334'
        mac = '00:00:fa:11:22:34'
        net_uuid = '0000-aaaa-bbbb-cccc'
        segmentation_id = 10001
        status = 'down'
        oui = None
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (del_flow), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.get_ofport_name',
                       return_value='test_port'), \
            mock.patch.object(self.ovs_vdp.lldpad_info,
                              'send_vdp_vnic_down') as vnic_down:
            self.ovs_vdp.local_vlan_map[net_uuid] = ovs_vdp.LocalVlan(10, (
                segmentation_id))
            lvm = self.ovs_vdp.local_vlan_map[net_uuid]
            lvm.lvid = 10
            lvm.port_uuid_list[port_uuid] = port_uuid
            self.ovs_vdp.local_vlan_map[net_uuid].late_binding_vlan = 500
            parent = mock.MagicMock()
            parent.attach_mock(vnic_down, 'send_vdp_vnic_down')
            parent.attach_mock(del_flow, 'delete_flows')
            self.ovs_vdp.send_vdp_port_event(port_uuid, mac, net_uuid,
                                             segmentation_id, status, oui)
        expected_calls = [mock.call.send_vdp_vnic_down(port_uuid=port_uuid,
                                                       vsiid=port_uuid,
                                                       gid=segmentation_id,
                                                       mac=mac, vlan=500,
                                                       oui=oui),
                          mock.call.delete_flows(
                              in_port=self.ovs_vdp.phy_peer_port_num,
                              dl_vlan=10),
                          mock.call.delete_flows(
                              in_port=self.ovs_vdp.int_peer_port_num,
                              dl_vlan=500)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_vdp_port_event(self):
        '''
        Routine the calls the other new port and existing port test routines
        '''
        self._test_vdp_port_event_new()
        self._test_vdp_port_event_exist()

    def test_vdp_port_event_down(self):
        '''Routine the calls the port down test '''
        self._test_vdp_port_event_down()

    def test_vdp_port_event_down_valid_vlan(self):
        """Test the case for a vnic port down for a network.

        This is to test the case when there are more than one vNic for a
        network with a valid VLAN. Flows should not be removed.
        """
        status = 'down'
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (del_flow), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.get_ofport_name',
                       return_value='test_port'), \
            mock.patch.object(self.ovs_vdp.lldpad_info,
                              'send_vdp_vnic_down') as vnic_down:
            self.ovs_vdp.local_vlan_map[self.net_uuid] = ovs_vdp.LocalVlan(
                self.lvid, self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            lvm.set_port_uuid(self.port_uuid_1, self.exist_vdp_vlan, None)
            self.ovs_vdp.local_vlan_map[
                self.net_uuid].late_binding_vlan = self.exist_vdp_vlan
            parent = mock.MagicMock()
            parent.attach_mock(vnic_down, 'send_vdp_vnic_down')
            parent.attach_mock(del_flow, 'delete_flows')
            self.ovs_vdp.send_vdp_port_event(
                self.port_uuid, self.mac, self.net_uuid,
                self.segmentation_id, status, self.oui)
        expected_calls = [mock.call.send_vdp_vnic_down(
            port_uuid=self.port_uuid, vsiid=self.port_uuid,
            gid=self.segmentation_id, mac=self.mac, vlan=self.exist_vdp_vlan,
            oui=self.oui)]
        parent.assert_has_calls(expected_calls, any_order=False)
        del_flow.assert_not_called()

    def test_vdp_port_event_down_no_valid_vlan(self):
        """Test the case for a vnic port down for a network with no valid vlan.

       This is to test the case when there are more than one vNic for a
        network with no valid VLAN. Flows should be removed.
        """
        status = 'down'
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (del_flow), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.get_ofport_name',
                       return_value='test_port'), \
            mock.patch.object(self.ovs_vdp.lldpad_info,
                              'send_vdp_vnic_down') as vnic_down:
            self.ovs_vdp.local_vlan_map[self.net_uuid] = ovs_vdp.LocalVlan(
                self.lvid, self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            lvm.set_port_uuid(self.port_uuid_1, 0, None)
            self.ovs_vdp.local_vlan_map[
                self.net_uuid].late_binding_vlan = self.exist_vdp_vlan
            parent = mock.MagicMock()
            parent.attach_mock(vnic_down, 'send_vdp_vnic_down')
            parent.attach_mock(del_flow, 'delete_flows')
            self.ovs_vdp.send_vdp_port_event(
                self.port_uuid, self.mac, self.net_uuid,
                self.segmentation_id, status, self.oui)
        expected_calls = [
            mock.call.send_vdp_vnic_down(
                port_uuid=self.port_uuid, vsiid=self.port_uuid,
                gid=self.segmentation_id, mac=self.mac,
                vlan=self.exist_vdp_vlan, oui=self.oui),
            mock.call.delete_flows(in_port=self.ovs_vdp.phy_peer_port_num,
                                   dl_vlan=self.lvid),
            mock.call.delete_flows(in_port=self.ovs_vdp.int_peer_port_num,
                                   dl_vlan=self.exist_vdp_vlan)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_vdp_port_event_down_mismatched_vlans(self):
        """Test the case for a vnic port down for a network with mismatch vlan.

        This is to test the case when there are more than one vNic for a
        network with mismatched VLAN's. Stale Flows should be removed and new
        flows should be added.
        """
        old_vdp_vlan = 3001
        status = 'down'
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (del_flow), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.add_flow') as (add_flow), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.get_ofport_name',
                       return_value='test_port'), \
            mock.patch.object(self.ovs_vdp.lldpad_info,
                              'send_vdp_vnic_down') as vnic_down:
            self.ovs_vdp.local_vlan_map[self.net_uuid] = ovs_vdp.LocalVlan(
                self.lvid, self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            lvm.set_port_uuid(self.port_uuid_1, old_vdp_vlan, None)
            self.ovs_vdp.local_vlan_map[
                self.net_uuid].late_binding_vlan = self.exist_vdp_vlan
            parent = mock.MagicMock()
            parent.attach_mock(vnic_down, 'send_vdp_vnic_down')
            parent.attach_mock(del_flow, 'delete_flows')
            parent.attach_mock(add_flow, 'add_flow')
            self.ovs_vdp.send_vdp_port_event(
                self.port_uuid, self.mac, self.net_uuid,
                self.segmentation_id, status, self.oui)
        expected_calls = [
            mock.call.send_vdp_vnic_down(
                port_uuid=self.port_uuid, vsiid=self.port_uuid,
                gid=self.segmentation_id, mac=self.mac,
                vlan=self.exist_vdp_vlan, oui=self.oui),
            mock.call.delete_flows(in_port=self.ovs_vdp.phy_peer_port_num,
                                   dl_vlan=self.lvid),
            mock.call.delete_flows(in_port=self.ovs_vdp.int_peer_port_num,
                                   dl_vlan=self.exist_vdp_vlan),
            mock.call.add_flow(
                priority=4, in_port=self.ovs_vdp.phy_peer_port_num,
                dl_vlan=self.lvid,
                actions="mod_vlan_vid:%s,normal" % old_vdp_vlan),
            mock.call.add_flow(
                priority=3, in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=old_vdp_vlan,
                actions="mod_vlan_vid:%s,normal" % self.lvid)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_vdp_vlan_change_rem_add(self):
        """Testing the VDP VLAN change for a remove/add flow case. """
        new_vdp_vlan = 3006
        ovs_cb_data = {'port_uuid': self.port_uuid, 'net_uuid': self.net_uuid}
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (ovs_br_del):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_del, 'delete_flows')
            parent.attach_mock(self.ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            self.ovs_vdp.vdp_vlan_change(ovs_cb_data, new_vdp_vlan, None)
        expected_calls = [
            mock.call.delete_flows(
                in_port=self.ovs_vdp.phy_peer_port_num, dl_vlan=self.lvid),
            mock.call.delete_flows(
                in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=self.exist_vdp_vlan),
            mock.call.add_flow(
                priority=4, in_port=self.ovs_vdp.phy_peer_port_num,
                dl_vlan=self.lvid,
                actions="mod_vlan_vid:%s,normal" % new_vdp_vlan),
            mock.call.add_flow(
                priority=3, in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=new_vdp_vlan,
                actions="mod_vlan_vid:%s,normal" % self.lvid)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_vdp_vlan_change_rem(self):
        """Testing the VDP VLAN change for a remove flow case. """
        new_vdp_vlan = 0
        ovs_cb_data = {'port_uuid': self.port_uuid, 'net_uuid': self.net_uuid}
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.delete_flows') as (ovs_br_del):
            parent = mock.MagicMock()
            parent.reset_mock(self.ovs_br_add)
            parent.attach_mock(ovs_br_del, 'delete_flows')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            self.ovs_vdp.vdp_vlan_change(ovs_cb_data, new_vdp_vlan, None)
        expected_calls = [
            mock.call.delete_flows(
                in_port=self.ovs_vdp.phy_peer_port_num, dl_vlan=self.lvid),
            mock.call.delete_flows(
                in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=self.exist_vdp_vlan)]
        parent.assert_has_calls(expected_calls, any_order=False)
        self.ovs_br_add.assert_not_called()

    def test_vdp_vlan_change_multiple_vnics_norem(self):
        """Testing the VDP VLAN change for multiple vnic's.

        This is for the case when there are multiple vNics for the same
        network and for one vNic a VDP VLAN of 0, is returned. So flow
        should not be deleted.
        """
        new_vdp_vlan = 0
        ovs_cb_data = {'port_uuid': self.port_uuid, 'net_uuid': self.net_uuid}
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.add_flow') as (ovs_br_add), \
            mock.patch('networking_cisco.apps.saf.common.'
                       'dfa_sys_lib.OVSBridge.delete_flows') as (ovs_br_del):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_del, 'delete_flows')
            parent.attach_mock(ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            lvm.set_port_uuid(self.port_uuid_1, self.exist_vdp_vlan, None)
            self.ovs_vdp.vdp_vlan_change(ovs_cb_data, new_vdp_vlan, None)
        ovs_br_add.assert_not_called()
        ovs_br_del.assert_not_called()

    def test_flow_check_handler_ext_flows_missing(self):
        """Testing the flow check handler for external bridge.

        Flow is missing for external bridge.
        """
        integ_flow = "NXST_FLOW reply (xid=0x4):\n" \
                     "cookie=0x0, duration=460654.804s, table=0, "\
                     "n_packets=3885, n_bytes=248632, idle_age=71, "\
                     "hard_age=65534, priority=3,in_port=1,dl_vlan=10 "\
                     "actions=mod_vlan_vid:3005,NORMAL\n" \
                     "cookie=0x9feb720beeec4ab9, duration=777857.315s, "\
                     "table=0, n_packets=32387, n_bytes=2186200, idle_age=4,"\
                     " hard_age=65534, priority=2,in_port=1 actions=drop"
        ext_flow = "NXST_FLOW reply (xid=0x4):"
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.add_flow') as (ovs_br_add), \
            mock.patch.object(self.ovs_vdp.integ_br_obj, 'dump_flows_for',
                              return_value=integ_flow), \
            mock.patch.object(self.ovs_vdp.ext_br_obj, 'dump_flows_for',
                              return_value=ext_flow):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            self.ovs_vdp._flow_check_handler()
        expected_calls = [
            mock.call.add_flow(
                priority=4, in_port=self.ovs_vdp.phy_peer_port_num,
                dl_vlan=self.lvid,
                actions="mod_vlan_vid:%s,normal" % self.exist_vdp_vlan),
            mock.call.add_flow(
                priority=3, in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=self.exist_vdp_vlan,
                actions="mod_vlan_vid:%s,normal" % self.lvid)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_flow_check_handler_integ_flows_missing(self):
        """Testing the flow check handler for integ bridge.

        Flow is missing for integration bridge.
        """
        ext_flow = "NXST_FLOW reply (xid=0x4):\n" \
                   "cookie=0x0, duration=460654.804s, table=0, "\
                   "n_packets=3885, n_bytes=248632, idle_age=71, "\
                   "hard_age=65534, priority=3,in_port=1,dl_vlan=3005 "\
                   "actions=mod_vlan_vid:10,NORMAL\n" \
                   "cookie=0x9feb720beeec4ab9, duration=777857.315s, "\
                   "table=0, n_packets=32387, n_bytes=2186200, idle_age=4,"\
                   " hard_age=65534, priority=2,in_port=1 actions=drop"
        integ_flow = "NXST_FLOW reply (xid=0x4):"
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.add_flow') as (ovs_br_add), \
            mock.patch.object(self.ovs_vdp.integ_br_obj, 'dump_flows_for',
                              return_value=integ_flow), \
            mock.patch.object(self.ovs_vdp.ext_br_obj, 'dump_flows_for',
                              return_value=ext_flow):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            self.ovs_vdp._flow_check_handler()
        expected_calls = [
            mock.call.add_flow(
                priority=4, in_port=self.ovs_vdp.phy_peer_port_num,
                dl_vlan=self.lvid,
                actions="mod_vlan_vid:%s,normal" % self.exist_vdp_vlan),
            mock.call.add_flow(
                priority=3, in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=self.exist_vdp_vlan,
                actions="mod_vlan_vid:%s,normal" % self.lvid)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_flow_check_handler_both_flows_missing(self):
        """Testing the flow check handler for both bridges.

        Flow is missing for both bridges.
        """
        ext_flow = "NXST_FLOW reply (xid=0x4):"
        integ_flow = "NXST_FLOW reply (xid=0x4):"
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.add_flow') as (ovs_br_add), \
            mock.patch.object(self.ovs_vdp.integ_br_obj, 'dump_flows_for',
                              return_value=integ_flow), \
            mock.patch.object(self.ovs_vdp.ext_br_obj, 'dump_flows_for',
                              return_value=ext_flow):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = self.exist_vdp_vlan
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, self.exist_vdp_vlan, None)
            self.ovs_vdp._flow_check_handler()
        expected_calls = [
            mock.call.add_flow(
                priority=4, in_port=self.ovs_vdp.phy_peer_port_num,
                dl_vlan=self.lvid,
                actions="mod_vlan_vid:%s,normal" % self.exist_vdp_vlan),
            mock.call.add_flow(
                priority=3, in_port=self.ovs_vdp.int_peer_port_num,
                dl_vlan=self.exist_vdp_vlan,
                actions="mod_vlan_vid:%s,normal" % self.lvid)]
        parent.assert_has_calls(expected_calls, any_order=False)

    def test_flow_check_handler_no_flows_missing(self):
        """Testing the flow check handler for both bridges.

        No flows are missing in both the bridges.
        """
        new_vdp_vlan = 3005
        ext_flow = "NXST_FLOW reply (xid=0x4):\n" \
                   "cookie=0x0, duration=460654.804s, table=0, "\
                   "n_packets=3885, n_bytes=248632, idle_age=71, "\
                   "hard_age=65534, priority=3,in_port=1,dl_vlan=10 "\
                   "actions=mod_vlan_vid:3005,NORMAL\n" \
                   "cookie=0x9feb720beeec4ab9, duration=777857.315s, "\
                   "table=0, n_packets=32387, n_bytes=2186200, idle_age=4,"\
                   " hard_age=65534, priority=2,in_port=1 actions=drop"
        integ_flow = "NXST_FLOW reply (xid=0x4):\n" \
                     "cookie=0x0, duration=460654.804s, table=0, "\
                     "n_packets=3885, n_bytes=248632, idle_age=71, "\
                     "hard_age=65534, priority=3,in_port=1,dl_vlan=3005 "\
                     "actions=mod_vlan_vid:10,NORMAL\n" \
                     "cookie=0x9feb720beeec4ab9, duration=777857.315s, "\
                     "table=0, n_packets=32387, n_bytes=2186200, idle_age=4,"\
                     " hard_age=65534, priority=2,in_port=1 actions=drop"
        with mock.patch('networking_cisco.apps.saf.common.'
                        'dfa_sys_lib.OVSBridge.add_flow') as (ovs_br_add), \
            mock.patch.object(self.ovs_vdp.integ_br_obj, 'dump_flows_for',
                              return_value=integ_flow), \
            mock.patch.object(self.ovs_vdp.ext_br_obj, 'dump_flows_for',
                              return_value=ext_flow):
            parent = mock.MagicMock()
            parent.attach_mock(ovs_br_add, 'add_flow')
            self.ovs_vdp.local_vlan_map[
                self.net_uuid] = ovs_vdp.LocalVlan(self.lvid,
                                                   self.segmentation_id)
            lvm = self.ovs_vdp.local_vlan_map[self.net_uuid]
            lvm.late_binding_vlan = new_vdp_vlan
            lvm.lvid = self.lvid
            lvm.set_port_uuid(self.port_uuid, new_vdp_vlan, None)
            self.ovs_vdp._flow_check_handler()
        ovs_br_add.assert_not_called()

    def test_populate_cache(self):
        """Test the populate cache function. """
        self.ovs_vdp.pop_local_cache(
            self.port_uuid, self.mac, self.net_uuid, self.lvid,
            self.exist_vdp_vlan, self.segmentation_id)
        rvlan = self.ovs_vdp.local_vlan_map[self.net_uuid].get_portid_vlan(
            self.port_uuid)
        fail_reason = self.ovs_vdp.local_vlan_map[
            self.net_uuid].get_portid_fail_reason(self.port_uuid)
        valid_vlan = self.ovs_vdp.local_vlan_map[
            self.net_uuid].any_valid_vlan()
        self.assertEqual(
            self.lvid, self.ovs_vdp.local_vlan_map[self.net_uuid].lvid)
        self.assertEqual(
            self.exist_vdp_vlan,
            self.ovs_vdp.local_vlan_map[self.net_uuid].late_binding_vlan)
        self.assertEqual(
            False, self.ovs_vdp.local_vlan_map[self.net_uuid].vdp_nego_req)
        self.assertEqual(self.exist_vdp_vlan, rvlan)
        self.assertIsNone(fail_reason)
        self.assertTrue(valid_vlan)



# Copyright (c) 2017 Cisco Systems, Inc.
# All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from oslo_config import cfg

from neutron.tests import base

from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import ml2_api as api

from networking_cisco.ml2_drivers.nexus import (
    exceptions as excep)
from networking_cisco.ml2_drivers.nexus import (
    nexus_restapi_network_driver as rest_driver)
from networking_cisco.ml2_drivers.nexus.extensions import (
    cisco_providernet)
from networking_cisco.ml2_drivers.nexus import mech_cisco_nexus
from networking_cisco.ml2_drivers.nexus import nexus_db_v2
from networking_cisco.ml2_drivers.nexus import nexus_models_v2
from neutron.tests.unit import testlib_api

NETWORK_ID = 'network_id'
VLAN_ID = 'vlan_id'
DEVICE_ID = 'device_id'
HOST_ID = 'host_id'
PORT_ID = 'port_id'
IP_ADDR = 'ipaddr'
INTF_TYPE = 'intf_type'
NEXUS_PORT = 'nexus_port'
IS_NATIVE = True
NO_VNI = 0
NETWORK = {'id': NETWORK_ID,
           'is_provider_network': True,
           api.NETWORK_TYPE: 'vlan',
           api.SEGMENTATION_ID: VLAN_ID,
           bc.providernet.SEGMENTATION_ID: VLAN_ID}
PORT = {'device_id': DEVICE_ID,
        bc.portbindings.VNIC_TYPE: 'normal',
        bc.portbindings.HOST_ID: HOST_ID}


class TestCiscoNexusProvider(testlib_api.SqlTestCase):
    """Test the provider network code added to the cisco nexus MD."""

    def setUp(self):
        super(TestCiscoNexusProvider, self).setUp()
        self._nexus_md = mech_cisco_nexus.CiscoNexusMechanismDriver()
        self._nexus_md._get_port_uuid = mock.Mock(return_value='test_uuid')
        self._func = mock.Mock()
        self.context = mock.Mock()
        self.context.current = NETWORK

    def test_create_network(self):
        self._nexus_md.create_network_precommit(self.context)

        self.assertTrue(nexus_db_v2.is_provider_network(NETWORK_ID))
        self.assertTrue(nexus_db_v2.is_provider_vlan(VLAN_ID))

    def test_create_network_no_provider(self):
        NETWORK_NO_PROVIDER = NETWORK.copy()
        del NETWORK_NO_PROVIDER['is_provider_network']
        self.context.current = NETWORK_NO_PROVIDER
        self._nexus_md.create_network_precommit(self.context)

        self.assertFalse(nexus_db_v2.is_provider_network(NETWORK_ID))
        self.assertFalse(nexus_db_v2.is_provider_vlan(VLAN_ID))

    def test_create_network_false_provider(self):
        NETWORK_FALSE_PROVIDER = NETWORK.copy()
        NETWORK_FALSE_PROVIDER['is_provider_network'] = False
        self.context.current = NETWORK_FALSE_PROVIDER
        self._nexus_md.create_network_precommit(self.context)

        self.assertFalse(nexus_db_v2.is_provider_network(NETWORK_ID))
        self.assertFalse(nexus_db_v2.is_provider_vlan(VLAN_ID))

    def test_delete_network(self):
        self._nexus_md.create_network_precommit(self.context)
        self._nexus_md.delete_network_postcommit(self.context)

        self.assertFalse(nexus_db_v2.is_provider_network(NETWORK_ID))
        self.assertFalse(nexus_db_v2.is_provider_vlan(VLAN_ID))

    def test_delete_network_no_id(self):
        mock_subport_get_object = mock.patch.object(
            nexus_db_v2, 'delete_provider_network').start()
        self._nexus_md.delete_network_postcommit(self.context)

        self.assertFalse(mock_subport_get_object.call_count)

    def test_port_action_vlan_provider(self):
        self._nexus_md.create_network_precommit(self.context)
        self._nexus_md._port_action_vlan(PORT, NETWORK, self._func, 0)

        self._func.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, mock.ANY, mock.ANY, True)

    def test_port_action_vlan_no_provider(self):
        self._nexus_md._port_action_vlan(PORT, NETWORK, self._func, 0)

        self._func.assert_called_once_with(
            mock.ANY, mock.ANY, mock.ANY, mock.ANY, mock.ANY, False)


class TestCiscoNexusProviderExtension(base.BaseTestCase):
    """Test the provider network extension class used by the cisco nexus MD."""

    def setUp(self):
        super(TestCiscoNexusProviderExtension, self).setUp()
        self._context = mock.Mock()
        self._data = {}
        self._result = {}
        self._provider_net_driver = cisco_providernet.CiscoProviderNetDriver()

    def test_extension_alias(self):
        self.assertTrue(self._provider_net_driver.extension_alias ==
                        'provider')

    def test_create_network_vlan(self):
        self._data[bc.providernet.SEGMENTATION_ID] = VLAN_ID
        self._provider_net_driver.process_create_network(
            self._context, self._data, self._result)

        self.assertTrue(self._result['is_provider_network'])

    def test_create_network_no_vlan(self):
        self._provider_net_driver.process_create_network(
            self._context, self._data, self._result)

        self.assertFalse(self._result.get('is_provider_network'))

    def test_create_network_none_vlan(self):
        self._data[bc.providernet.SEGMENTATION_ID] = None
        self._provider_net_driver.process_create_network(
            self._context, self._data, self._result)

        self.assertFalse(self._result.get('is_provider_network'))


class TestCiscoNexusProviderConfiguration(base.BaseTestCase):
    """Test the provider network configuration used by the cisco nexus MD."""

    def setUp(self):
        super(TestCiscoNexusProviderConfiguration, self).setUp()
        mock.patch.object(rest_driver.CiscoNexusRestapiDriver,
                          '__init__', return_value=None).start()
        self._nexus_md = mech_cisco_nexus.CiscoNexusMechanismDriver()
        self._nexus_md.set_switch_ip_and_active_state = mock.Mock()
        self._nexus_md.configure_next_batch_of_vlans = mock.Mock()
        self._nexus_md.driver = rest_driver.CiscoNexusRestapiDriver()
        self._nexus_md.driver.capture_and_print_timeshot = mock.Mock()
        self._get_active_host_connections_mock = mock.patch.object(
            mech_cisco_nexus.CiscoNexusMechanismDriver,
            '_get_active_port_connections',
            return_value=[[IP_ADDR, INTF_TYPE, NEXUS_PORT, IS_NATIVE,
                           None]]).start()
        self._save_switch_vlan_range_mock = mock.patch.object(
            mech_cisco_nexus.CiscoNexusMechanismDriver,
            '_save_switch_vlan_range').start()
        self._restore_port_binding_mock = mock.patch.object(
            mech_cisco_nexus.CiscoNexusMechanismDriver,
            '_restore_port_binding').start()
        self._create_and_trunk_vlan_mock = mock.patch.object(
            rest_driver.CiscoNexusRestapiDriver,
            'create_and_trunk_vlan').start()
        self._create_vlan_mock = mock.patch.object(
            rest_driver.CiscoNexusRestapiDriver,
            'create_vlan').start()
        self._send_enable_vlan_on_trunk_int_mock = mock.patch.object(
            rest_driver.CiscoNexusRestapiDriver,
            'send_enable_vlan_on_trunk_int').start()
        self._disable_vlan_on_trunk_int_mock = mock.patch.object(
            rest_driver.CiscoNexusRestapiDriver,
            'disable_vlan_on_trunk_int').start()
        self._get_nexusvlan_binding_mock = mock.patch.object(
            nexus_db_v2, 'get_nexusvlan_binding').start()
        self._is_provider_vlan_mock = mock.patch.object(
            nexus_db_v2, 'is_provider_vlan').start()
        mock.patch.object(
            nexus_db_v2, 'get_port_vlan_switch_binding',
            side_effect=excep.NexusPortBindingNotFound).start()
        mock.patch.object(
            mech_cisco_nexus.CiscoNexusMechanismDriver,
            '_delete_port_channel_resources').start()

    def _set_provider_configuration(self, auto_create, auto_trunk):
        cfg.CONF.set_override('provider_vlan_auto_create', auto_create,
                              'ml2_cisco')
        cfg.CONF.set_override('provider_vlan_auto_trunk', auto_trunk,
                              'ml2_cisco')

    def _test_pnet_configure(
            self, auto_create, auto_trunk, is_provider_vlan=True):
        self._set_provider_configuration(auto_create, auto_trunk)
        self._nexus_md._configure_port_binding(
            is_provider_vlan, None, IS_NATIVE, IP_ADDR, VLAN_ID, INTF_TYPE,
            NEXUS_PORT, NO_VNI)

    def _test_pnet_delete(
            self, auto_create, auto_trunk, is_provider_vlan=True):
        self._set_provider_configuration(auto_create, auto_trunk)
        self._nexus_md._delete_switch_entry(
            PORT, VLAN_ID, DEVICE_ID, HOST_ID, NO_VNI, is_provider_vlan)

    def _test_pnet_replay(
            self, auto_create, auto_trunk, is_provider_vlan=True):
        self._set_provider_configuration(auto_create, auto_trunk)
        self._is_provider_vlan_mock.return_value = is_provider_vlan
        port_bindings = nexus_models_v2.NexusPortBinding(
            port_id=PORT_ID, vlan_id=VLAN_ID, vni=NO_VNI, switch_ip=IP_ADDR,
            instance_id=0, is_native=False, channel_group=0)
        self._nexus_md.configure_switch_entries(IP_ADDR, [port_bindings])

    def test_pnet_configure_create_and_trunk(self):
        self._test_pnet_configure(auto_create=True, auto_trunk=True)

        self._create_and_trunk_vlan_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, NO_VNI, True)
        self.assertFalse(self._create_vlan_mock.call_count)
        self.assertFalse(self._send_enable_vlan_on_trunk_int_mock.call_count)

    def test_pnet_configure_create(self):
        self._test_pnet_configure(auto_create=True, auto_trunk=False)

        self.assertFalse(self._create_and_trunk_vlan_mock.call_count)
        self._create_vlan_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, NO_VNI)
        self.assertFalse(self._send_enable_vlan_on_trunk_int_mock.call_count)

    def test_pnet_configure_trunk(self):
        self._test_pnet_configure(auto_create=False, auto_trunk=True)

        self.assertFalse(self._create_and_trunk_vlan_mock.call_count)
        self.assertFalse(self._create_vlan_mock.call_count)
        self._send_enable_vlan_on_trunk_int_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, True)

    def test_pnet_configure_not_providernet(self):
        self._test_pnet_configure(
            auto_create=False, auto_trunk=False, is_provider_vlan=False)

        self._create_and_trunk_vlan_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, NO_VNI, True)
        self.assertFalse(self._create_vlan_mock.call_count)
        self.assertFalse(self._send_enable_vlan_on_trunk_int_mock.call_count)

    def test_pnet_delete_create_and_trunk(self):
        self._test_pnet_delete(auto_create=True, auto_trunk=True)

        self._disable_vlan_on_trunk_int_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, IS_NATIVE)
        self._get_nexusvlan_binding_mock.assert_called_once_with(
            VLAN_ID, IP_ADDR)

    def test_pnet_delete_trunk(self):
        self._test_pnet_delete(auto_create=False, auto_trunk=True)

        self._disable_vlan_on_trunk_int_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, IS_NATIVE)
        self.assertFalse(self._get_nexusvlan_binding_mock.call_count)

    def test_pnet_delete_create(self):
        self._test_pnet_delete(auto_create=True, auto_trunk=False)

        self.assertFalse(self._disable_vlan_on_trunk_int_mock.call_count)
        self._get_nexusvlan_binding_mock.assert_called_once_with(
            VLAN_ID, IP_ADDR)

    def test_pnet_delete_not_providernet(self):
        self._test_pnet_delete(
            auto_create=False, auto_trunk=False, is_provider_vlan=False)

        self._disable_vlan_on_trunk_int_mock.assert_called_once_with(
            IP_ADDR, VLAN_ID, INTF_TYPE, NEXUS_PORT, IS_NATIVE)
        self._get_nexusvlan_binding_mock.assert_called_once_with(
            VLAN_ID, IP_ADDR)

    def test_pnet_replay_not_providernet(self):
        self._test_pnet_replay(
            auto_create=False, auto_trunk=False, is_provider_vlan=False)

        self._save_switch_vlan_range_mock.assert_called_once_with(
            IP_ADDR, [(VLAN_ID, NO_VNI)])
        self._restore_port_binding_mock.assert_called_once_with(
            IP_ADDR, set([VLAN_ID]), mock.ANY, mock.ANY)

    def test_pnet_replay_providernet_create_and_trunk(self):
        self._test_pnet_replay(
            auto_create=True, auto_trunk=True, is_provider_vlan=True)

        self._save_switch_vlan_range_mock.assert_called_once_with(
            IP_ADDR, [(VLAN_ID, NO_VNI)])
        self._restore_port_binding_mock.assert_called_once_with(
            IP_ADDR, set([VLAN_ID]), mock.ANY, mock.ANY)

    def test_pnet_replay_providernet_create(self):
        self._test_pnet_replay(
            auto_create=True, auto_trunk=False, is_provider_vlan=True)

        self._save_switch_vlan_range_mock.assert_called_once_with(
            IP_ADDR, [(VLAN_ID, NO_VNI)])
        self.assertFalse(self._restore_port_binding_mock.call_count)

    def test_pnet_replay_providernet_trunk(self):
        self._test_pnet_replay(
            auto_create=False, auto_trunk=True, is_provider_vlan=True)

        self.assertFalse(self._save_switch_vlan_range_mock.call_count)
        self._restore_port_binding_mock.assert_called_once_with(
            IP_ADDR, set([VLAN_ID]), mock.ANY, mock.ANY)



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import random

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils
import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm import exc
from sqlalchemy.orm import joinedload
from sqlalchemy.sql import expression as expr

from neutron.common import utils
from neutron.db import models_v2
from neutron_lib import exceptions as n_exc

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import l3_const
from networking_cisco.backwards_compatibility import l3_exceptions
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.common import utils as cisco_utils
from networking_cisco.plugins.cisco.extensions import ha
from networking_cisco.plugins.cisco.extensions import routerrole
from networking_cisco.plugins.cisco.extensions import routertype

LOG = logging.getLogger(__name__)


RouterPort = bc.RouterPort

HA_GROUP = 'group'
HA_PORT = 'ha_port'

MAX_VRRP_GROUPS = 4094
MAX_HSRP_GROUPS = 4094
MAX_GLBP_GROUPS = 1023

is_attr_set = bc.is_attr_set
ATTR_NOT_SPECIFIED = bc.constants.ATTR_NOT_SPECIFIED
EXTERNAL_GW_INFO = l3_const.EXTERNAL_GW_INFO
DEVICE_OWNER_ROUTER_GW = bc.constants.DEVICE_OWNER_ROUTER_GW
DEVICE_OWNER_ROUTER_INTF = bc.constants.DEVICE_OWNER_ROUTER_INTF
DEVICE_OWNER_ROUTER_HA_INTF = bc.constants.DEVICE_OWNER_ROUTER_HA_INTF
ROUTER_ROLE_HA_REDUNDANCY = cisco_constants.ROUTER_ROLE_HA_REDUNDANCY

DEFAULT_MASTER_PRIORITY = 100
PRIORITY_INCREASE_STEP = -3
REDUNDANCY_ROUTER_SUFFIX = '_HA_backup_'
DEFAULT_PING_INTERVAL = 5
PROBE_TARGET_OPT_NAME = 'default_probe_target'

LOOKUP_RETRIES = 5
RETRY_INTERVAL = 3

router_appliance_opts = [
    cfg.BoolOpt('ha_support_enabled', default=True,
                help=_("Enables high-availability support")),
    cfg.BoolOpt('ha_enabled_by_default', default=False,
                help=_("Enables high-availability functionality for Neutron "
                       "router even if user does not explicitly request it")),
    cfg.IntOpt('default_ha_redundancy_level', default=ha.MIN_REDUNDANCY_LEVEL,
               help=_("Default number of routers added for redundancy when "
                      "high-availability by VRRP, HSRP, or GLBP is used")),
    cfg.StrOpt('default_ha_mechanism', default=ha.HA_HSRP,
               help=_("Default mechanism used to implement "
                      "high-availability")),
    cfg.ListOpt('disabled_ha_mechanisms', default=[],
                help=_("List of administratively disabled high-availability "
                       "mechanisms (VRRP, HSRP, GBLP)")),
    cfg.BoolOpt('connectivity_probing_enabled_by_default', default=False,
                help=_("Enables connectivity probing for high-availability "
                       "even if user does not explicitly request it")),
    cfg.StrOpt('default_probe_target', default=None,
               help=_("Host that will be probe target for high-availability "
                      "connectivity probing if user does not specify it")),
    cfg.StrOpt('default_ping_interval', default=DEFAULT_PING_INTERVAL,
               help=_("Time (in seconds) between probes for high-availability "
                      "connectivity probing if user does not specify it")),
]

cfg.CONF.register_opts(router_appliance_opts, "ha")


class RouterHASetting(bc.model_base.BASEV2):
    """Represents HA settings for router visible to user."""
    __tablename__ = 'cisco_router_ha_settings'

    router_id = sa.Column(sa.String(36),
                          sa.ForeignKey('routers.id', ondelete='CASCADE'),
                          primary_key=True)
    router = orm.relationship(
        bc.Router,
        backref=orm.backref('ha_settings', cascade='all', uselist=False))
    # 'ha_type' can be 'VRRP', 'HSRP', or 'GLBP'
    ha_type = sa.Column(sa.String(255))
    # 'redundancy_level' is number of extra routers for redundancy
    redundancy_level = sa.Column(sa.Integer,
                                 server_default=str(ha.MIN_REDUNDANCY_LEVEL))
    # 'priority' is the priority used in VRRP, HSRP, and GLBP
    priority = sa.Column(sa.Integer)
    # 'probe_connectivity' is True if ICMP echo pinging is enabled
    probe_connectivity = sa.Column(sa.Boolean)
    # 'probe_target' is ip address of host that is probed
    probe_target = sa.Column(sa.String(64))
    # 'ping_interval' is the time between probes
    probe_interval = sa.Column(sa.Integer)
    # 'state' is the state of the user visible router: HA_ACTIVE or HA_STANDBY
    state = sa.Column(sa.Enum(ha.HA_ACTIVE, ha.HA_STANDBY, name='ha_states'),
                      default=ha.HA_ACTIVE, server_default=ha.HA_ACTIVE)


class RouterHAGroup(bc.model_base.BASEV2, bc.model_base.HasId,
                    bc.HasProject):
    """Represents an HA group as used in VRRP, HSRP, and GLBP."""
    __tablename__ = 'cisco_router_ha_groups'

    # 'ha_type' can be 'VRRP', 'HSRP', or 'GLBP'
    ha_type = sa.Column(sa.String(255))
    # 'group_identity'
    group_identity = sa.Column(sa.String(255))
    # 'ha_port_id' is id of port used for virtual IP address
    ha_port_id = sa.Column(sa.String(36),
                           sa.ForeignKey('ports.id', ondelete='CASCADE'),
                           primary_key=True)
    ha_port = orm.relationship(
        models_v2.Port,
        primaryjoin='Port.id==RouterHAGroup.ha_port_id')
    # 'extra_port_id' is id of port for user visible router's extra ip address
    extra_port_id = sa.Column(sa.String(36),
                              sa.ForeignKey('ports.id', ondelete='SET NULL'),
                              nullable=True)
    extra_port = orm.relationship(
        models_v2.Port,
        primaryjoin='Port.id==RouterHAGroup.extra_port_id')
    # 'subnet_id' is id of subnet that this HA group serves
    subnet_id = sa.Column(sa.String(36),
                          sa.ForeignKey('subnets.id', ondelete='CASCADE'),
                          primary_key=True)
    # 'user_router_id' is id of router visible to the user
    user_router_id = sa.Column(sa.String(36),
                               sa.ForeignKey('routers.id'))
    # 'timers_config' holds timer specific configurations
    timers_config = sa.Column(sa.String(255))
    # 'tracking_config' holds tracking object specific configurations
    tracking_config = sa.Column(sa.String(255))
    # 'other_config' holds other method specific configurations
    other_config = sa.Column(sa.String(255))


class RouterRedundancyBinding(bc.model_base.BASEV2):
    """Represents binding between an HA enabled router and its
    redundancy routers.
    """
    __tablename__ = 'cisco_router_redundancy_bindings'

    # 'redundancy_router_id' is id of a redundancy router
    redundancy_router_id = sa.Column(sa.String(36),
                                     sa.ForeignKey('routers.id',
                                                   ondelete='CASCADE'),
                                     primary_key=True)
    redundancy_router = orm.relationship(
        bc.Router,
        primaryjoin='Router.id==RouterRedundancyBinding.redundancy_router_id',
        backref=orm.backref('redundancy_binding', cascade='save-update, merge',
                            passive_deletes='all', uselist=False))
    # 'priority' is the priority used in VRRP, HSRP, and GLBP
    priority = sa.Column(sa.Integer)
    # 'state' is the state of the redundancy router: HA_ACTIVE or HA_STANDBY
    state = sa.Column(sa.Enum(ha.HA_ACTIVE, ha.HA_STANDBY, name='ha_states'),
                      default=ha.HA_STANDBY, server_default=ha.HA_STANDBY)
    # 'user_router_id' is id of router visible to the user
    user_router_id = sa.Column(sa.String(36),
                               sa.ForeignKey('routers.id'))
    user_router = orm.relationship(
        bc.Router,
        primaryjoin='Router.id==RouterRedundancyBinding.user_router_id',
        backref=orm.backref('redundancy_bindings',
                            order_by=priority, cascade='all'))
    __mapper_args__ = {
        'confirm_deleted_rows': False
    }


class HA_db_mixin(object):
    """Mixin class to support VRRP, HSRP, and GLBP based HA for routing."""

    def _ensure_create_ha_compliant(self, router, router_type):
        """To be called in create_router() BEFORE router is created in DB."""
        details = router.pop(ha.DETAILS, {})
        if details == ATTR_NOT_SPECIFIED:
            details = {}
        res = {ha.ENABLED: router.pop(ha.ENABLED, ATTR_NOT_SPECIFIED),
               ha.DETAILS: details}

        if not is_attr_set(res[ha.ENABLED]):
            res[ha.ENABLED] = router_type['ha_enabled_by_default']
        if res[ha.ENABLED] and not cfg.CONF.ha.ha_support_enabled:
            raise ha.HADisabled()
        if not res[ha.ENABLED]:
            return res
        if not is_attr_set(details.get(ha.TYPE, ATTR_NOT_SPECIFIED)):
            details[ha.TYPE] = cfg.CONF.ha.default_ha_mechanism
        if details[ha.TYPE] in cfg.CONF.ha.disabled_ha_mechanisms:
            raise ha.HADisabledHAType(ha_type=details[ha.TYPE])
        if not is_attr_set(details.get(ha.REDUNDANCY_LEVEL,
                                       ATTR_NOT_SPECIFIED)):
            details[ha.REDUNDANCY_LEVEL] = (
                cfg.CONF.ha.default_ha_redundancy_level)
        if not is_attr_set(details.get(ha.PROBE_CONNECTIVITY,
                                       ATTR_NOT_SPECIFIED)):
            details[ha.PROBE_CONNECTIVITY] = (
                cfg.CONF.ha.connectivity_probing_enabled_by_default)
        if not is_attr_set(details.get(ha.PROBE_TARGET, ATTR_NOT_SPECIFIED)):
            details[ha.PROBE_TARGET] = cfg.CONF.ha.default_probe_target
        if not is_attr_set(details.get(ha.PROBE_INTERVAL, ATTR_NOT_SPECIFIED)):
            details[ha.PROBE_INTERVAL] = cfg.CONF.ha.default_ping_interval
        return res

    def _create_redundancy_routers(self, context, new_router, ha_settings,
                                   new_router_db, ports=None, expire_db=False):
        """To be called in create_router() AFTER router has been
        created in DB.
        """
        if (ha.ENABLED not in ha_settings or
                not ha_settings[ha.ENABLED]):
            new_router[ha.HA] = {ha.ENABLED: False}
            return
        ha_spec = ha_settings[ha.DETAILS]
        priority = ha_spec.get(ha.PRIORITY, DEFAULT_MASTER_PRIORITY)
        with context.session.begin(subtransactions=True):
            r_ha_s_db = RouterHASetting(
                router_id=new_router['id'],
                ha_type=ha_spec[ha.TYPE],
                redundancy_level=ha_spec[ha.REDUNDANCY_LEVEL],
                priority=priority,
                probe_connectivity=ha_spec[ha.PROBE_CONNECTIVITY],
                probe_target=ha_spec[ha.PROBE_TARGET],
                probe_interval=ha_spec[ha.PROBE_INTERVAL])
            context.session.add(r_ha_s_db)
        if r_ha_s_db.probe_connectivity and r_ha_s_db.probe_target is None:
            LOG.warning("Connectivity probing for high-availability is "
                        "enabled but probe target is not specified. Please"
                        " configure option \'default_probe_target\'.")
        e_context = context.elevated()
        if new_router_db.gw_port:
            # generate ha settings and extra port for router gateway (VIP) port
            gw_port = self._core_plugin._make_port_dict(new_router_db.gw_port)
            self._create_ha_group(e_context, new_router, gw_port, r_ha_s_db)
        self._add_redundancy_routers(e_context, 1,
                                     ha_spec[ha.REDUNDANCY_LEVEL] + 1,
                                     new_router, ports or [], r_ha_s_db)
        if expire_db:
            context.session.expire(new_router_db)
        self._extend_router_dict_ha(new_router, new_router_db)

    def _ensure_update_ha_compliant(self, router, current_router,
                                    r_hd_binding_db):
        """To be called in update_router() BEFORE router has been
        updated in DB.
        """
        if r_hd_binding_db.role == ROUTER_ROLE_HA_REDUNDANCY:
            return {ha.ENABLED: False}
        auto_enable_ha = r_hd_binding_db.router_type.ha_enabled_by_default
        requested_ha_details = router.pop(ha.DETAILS, {})
        # If ha_details are given then ha is assumed to be enabled even if
        # it is not explicitly specified or if auto_enable_ha says so.
        # Note that None is used to indicate that request did not include any
        # ha information was provided!
        requested_ha_enabled = router.pop(
            ha.ENABLED, True if requested_ha_details or auto_enable_ha is True
            else None)
        res = {}
        ha_currently_enabled = current_router.get(ha.ENABLED, False)
        # Note: must check for 'is True' as None implies attribute not given
        if requested_ha_enabled is True or ha_currently_enabled is True:
            if not cfg.CONF.ha.ha_support_enabled:
                raise ha.HADisabled()
            curr_ha_details = current_router.get(ha.DETAILS, {})
            if ha.TYPE in requested_ha_details:
                requested_ha_type = requested_ha_details[ha.TYPE]
                if (ha.TYPE in curr_ha_details and
                        requested_ha_type != curr_ha_details[ha.TYPE]):
                    raise ha.HATypeCannotBeChanged()
                elif requested_ha_type in cfg.CONF.ha.disabled_ha_mechanisms:
                    raise ha.HADisabledHAType(ha_type=requested_ha_type)
        if requested_ha_enabled:
            res[ha.ENABLED] = requested_ha_enabled
            if requested_ha_details:
                res[ha.DETAILS] = requested_ha_details
        elif requested_ha_enabled is False:
            res[ha.ENABLED] = False
        return res

    def _teardown_redundancy_router_gw_connectivity(self, context, router,
                                                    router_db,
                                                    plugging_driver):
        """To be called in update_router() if the router gateway is to change
        BEFORE router has been updated in DB .
        """
        if not router[ha.ENABLED]:
            # No HA currently enabled so we're done
            return
        e_context = context.elevated()
        # since gateway is about to change the ha group for the current gateway
        # is removed, a new one will be created later
        self._delete_ha_group(e_context, router_db.gw_port_id)
        # teardown connectivity for the gw ports on the redundancy routers
        # and remove those ports as new ones will be created later
        rr_ids = []
        for r_b_db in router_db.redundancy_bindings:
            if plugging_driver is not None:
                plugging_driver.teardown_logical_port_connectivity(
                    e_context, r_b_db.redundancy_router.gw_port,
                    r_b_db.redundancy_router.hosting_info.hosting_device_id)
            self._update_router_no_notify(
                e_context, r_b_db.redundancy_router_id,
                {'router': {EXTERNAL_GW_INFO: None, ha.ENABLED: False}})
            rr_ids.append(r_b_db.redundancy_router_id)
        self.notify_routers_updated(e_context, rr_ids)

    def _update_redundancy_routers(self, context, updated_router,
                                   update_specification, requested_ha_settings,
                                   updated_router_db, gateway_changed):
        """To be called in update_router() AFTER router has been
        updated in DB.
        """
        router_requested = update_specification['router']
        ha_settings_db = updated_router_db.ha_settings
        ha_enabled_requested = requested_ha_settings.get(ha.ENABLED, False)
        if not (updated_router[ha.ENABLED] or ha_enabled_requested):
            # No HA currently enabled and no HA requested so we're done
            return
        # The redundancy routers need interfaces on the same networks as the
        # user visible router.
        ports = self._get_router_interfaces(updated_router_db)
        e_context = context.elevated()
        if not updated_router[ha.ENABLED] and ha_enabled_requested:
            # No HA currently enabled but HA requested
            router_requested.update(requested_ha_settings)
            router_requested[EXTERNAL_GW_INFO] = (
                updated_router[EXTERNAL_GW_INFO])
            requested_ha_settings = self._ensure_create_ha_compliant(
                router_requested, updated_router_db.hosting_info.router_type)
            self._create_redundancy_routers(
                e_context, updated_router, requested_ha_settings,
                updated_router_db, ports, expire_db=True)
            return
        rr_ids = self._get_redundancy_router_ids(context, updated_router['id'])
        ha_details_update_spec = requested_ha_settings.get(ha.DETAILS)
        if (updated_router[ha.ENABLED] and not requested_ha_settings.get(
                ha.ENABLED, updated_router[ha.ENABLED])):
            # HA currently enabled but HA disable requested
            # delete ha settings and extra port for gateway (VIP) port
            self._delete_ha_group(e_context, updated_router_db.gw_port_id)
            self._remove_redundancy_routers(e_context, rr_ids, ports, True)
            with context.session.begin(subtransactions=True):
                context.session.delete(ha_settings_db)
        elif ha_details_update_spec:
            # HA currently enabled and HA setting update (other than
            # disable HA) requested
            old_redundancy_level = ha_settings_db.redundancy_level
            ha_settings_db.update(ha_details_update_spec)
            diff = (ha_details_update_spec.get(ha.REDUNDANCY_LEVEL,
                                               old_redundancy_level) -
                    old_redundancy_level)
            with context.session.begin(subtransactions=True):
                context.session.add(ha_settings_db)
            if diff < 0:
                # Remove -diff redundancy routers
                #TODO(bobmel): Ensure currently active router is excluded
                to_remove = rr_ids[len(rr_ids) + diff:]
                rr_ids = rr_ids[:len(rr_ids) + diff]
                self._remove_redundancy_routers(e_context, to_remove, ports)
            elif diff > 0:
                # Add diff redundancy routers
                start = old_redundancy_level + 1
                stop = start + diff
                self._add_redundancy_routers(e_context, start, stop,
                                             updated_router, ports,
                                             ha_settings_db, False)
            if gateway_changed is True:
                self._change_ha_for_gateway(e_context, updated_router,
                                            updated_router_db, ha_settings_db,
                                            router_requested, expire=True)
            else:
                # Notify redundancy routers about changes
                self.notify_routers_updated(e_context, rr_ids)

        elif gateway_changed is True:
            # HA currently enabled (and to remain so) nor any HA setting update
            # and gateway has changed
            self._change_ha_for_gateway(e_context, updated_router,
                                        updated_router_db, ha_settings_db,
                                        router_requested)
        # pick up updates to other attributes where it makes sense
        # and push - right now it is only admin_state_up.
        other_updates_spec = {'router': {}}
        if 'admin_state_up' in update_specification['router']:
            other_updates_spec['router']['admin_state_up'] = (
                update_specification['router']['admin_state_up'])
        if 'name' in update_specification['router']:
            other_updates_spec['router']['name'] = (
                update_specification['router']['name'])
        if (other_updates_spec['router'] or
            'routes' in update_specification['router']):
            self._process_other_router_updates(e_context, updated_router_db,
                                               other_updates_spec)
        # Ensure we get latest state from DB
        context.session.expire(updated_router_db)
        self._extend_router_dict_ha(updated_router, updated_router_db)

    def _change_ha_for_gateway(self, context, router, router_db,
                               ha_settings_db, update_spec, expire=False):
        # generate ha settings and extra port for router gateway VIP
        if router_db.gw_port is None:
            # gateway was removed and since gw's of redundancy routers are
            # removed by _teardown_redundancy_router_gw_connectivity we're done
            return
        gw_port = self._core_plugin._make_port_dict(router_db.gw_port)
        self._create_ha_group(context, router, gw_port, ha_settings_db)
        if expire is True:
            context.session.expire(router_db)
        # Now add gw to redundancy routers
        rr_ids = []
        for r_b_db in router_db.redundancy_bindings:
            spec = {EXTERNAL_GW_INFO: copy.deepcopy(router[EXTERNAL_GW_INFO])}
            if spec[EXTERNAL_GW_INFO]['external_fixed_ips']:
                # Ensure ip addresses are not specified as they cannot be
                # same as visible router's ip addresses.
                for e_fixed_ip in spec[EXTERNAL_GW_INFO]['external_fixed_ips']:
                    e_fixed_ip.pop('ip_address', None)
            #spec[EXTERNAL_GW_INFO].pop('external_fixed_ips', None)
            spec[ha.ENABLED] = False
            self._update_router_no_notify(
                context, r_b_db.redundancy_router_id, {'router': spec})
            rr_ids.append(r_b_db.redundancy_router_id)
        self.notify_routers_updated(context, rr_ids)

    def _process_other_router_updates(self, context, router_db, update_spec):
        rr_ids = []
        new_name_stub = update_spec['router'].get('name')
        for r_b_db in router_db.redundancy_bindings:
            if update_spec['router']:
                if new_name_stub is not None:
                    idx = r_b_db.redundancy_router.name.split('_')[-1]
                    update_spec['router']['name'] = (
                        new_name_stub + REDUNDANCY_ROUTER_SUFFIX + idx)
                update_spec['router'][ha.ENABLED] = False
                self._update_router_no_notify(
                    context, r_b_db.redundancy_router_id, update_spec)
            rr_ids.append(r_b_db.redundancy_router_id)
        self.notify_routers_updated(context, rr_ids)

    def _add_redundancy_routers(self, context, start_index, stop_index,
                                user_visible_router, ports=None,
                                ha_settings_db=None, create_ha_group=True):
        """Creates a redundancy router and its interfaces on
        the specified subnets.
        """
        priority = (DEFAULT_MASTER_PRIORITY +
                    (start_index - 1) * PRIORITY_INCREASE_STEP)
        r = copy.deepcopy(user_visible_router)
        # No tenant_id so redundancy routers are hidden from user
        r['tenant_id'] = ''
        name = r['name']
        redundancy_r_ids = []
        for i in range(start_index, stop_index):
            del r['id']
            # We don't replicate the user visible router's routes, instead
            # they are populated to redundancy routers for get router(s) ops
            r.pop('routes', None)
            # Redundancy routers will never have a route spec themselves
            # The redundancy routers must have HA disabled
            r[ha.ENABLED] = False
            r['name'] = name + REDUNDANCY_ROUTER_SUFFIX + str(i)
            # set role so that purpose of this router can be easily determined
            r[routerrole.ROUTER_ROLE_ATTR] = ROUTER_ROLE_HA_REDUNDANCY
            gw_info = r[EXTERNAL_GW_INFO]
            if gw_info and gw_info['external_fixed_ips']:
                # Ensure ip addresses are not specified as they cannot be
                # same as visible router's ip addresses.
                for e_fixed_ip in gw_info['external_fixed_ips']:
                    e_fixed_ip.pop('ip_address', None)
            r = self.create_router(context, {'router': r})
            LOG.debug("Created redundancy router %(index)d with router id "
                      "%(r_id)s", {'index': i, 'r_id': r['id']})
            priority += PRIORITY_INCREASE_STEP
            r_b_b = RouterRedundancyBinding(
                redundancy_router_id=r['id'],
                priority=priority,
                user_router_id=user_visible_router['id'])
            context.session.add(r_b_b)
            redundancy_r_ids.append(r['id'])
        for port_db in ports or []:
            port = self._core_plugin._make_port_dict(port_db)
            self._add_redundancy_router_interfaces(
                context, user_visible_router, None, port,
                redundancy_r_ids, ha_settings_db, create_ha_group)

    def _remove_redundancy_routers(self, context, router_ids, ports,
                                   delete_ha_groups=False):
        """Deletes all interfaces of the specified redundancy routers
        and then the redundancy routers themselves.
        """
        subnets_info = [{'subnet_id': port['fixed_ips'][0]['subnet_id']}
                        for port in ports]
        for r_id in router_ids:
            for i in range(len(subnets_info)):
                self.remove_router_interface(context, r_id, subnets_info[i])
                LOG.debug("Removed interface on %(s_id)s to redundancy router "
                          "with %(r_id)s",
                          {'s_id': ports[i]['network_id'], 'r_id': r_id})
                # There is only one ha group per network so only delete once
                if delete_ha_groups and r_id == router_ids[0]:
                    self._delete_ha_group(context, ports[i]['id'])
            self.delete_router(context, r_id)
            LOG.debug("Deleted redundancy router %s", r_id)

    def _get_router_interfaces(self, router_db,
                               port_type=DEVICE_OWNER_ROUTER_INTF):
        return [p['port'] for p in router_db.attached_ports if
                p['port_type'] == port_type]

    def _delete_redundancy_routers(self, context, router_db):
        """To be called in delete_router() BEFORE router has been
        deleted in DB. The router should have not interfaces.
        """
        e_context = context.elevated()
        for binding in router_db.redundancy_bindings:
            self.delete_router(e_context, binding.redundancy_router_id)
            LOG.debug("Deleted redundancy router %s",
                      binding.redundancy_router_id)
        if router_db.gw_port_id:
            # delete ha settings and extra port for gateway (VIP) port
            self._delete_ha_group(e_context, router_db.gw_port_id)

    def _add_redundancy_router_interfaces(self, context, router, itfc_info,
                                          new_port, redundancy_router_ids=None,
                                          ha_settings_db=None,
                                          create_ha_group=True):
        """To be called in add_router_interface() AFTER interface has been
        added to router in DB.
        """
        # There are essentially three cases where we add interface to a
        # redundancy router:
        # 1. HA is enabled on a user visible router that has one or more
        #    interfaces.
        # 2. Redundancy level is increased so one or more redundancy routers
        #    are added.
        # 3. An interface is added to a user visible router.
        #
        # For 1: An HA GROUP MUST BE CREATED and EXTRA PORTS MUST BE CREATED
        #        for each redundancy router. The id of extra port should be
        #        specified in the interface_info argument of the
        #        add_router_interface call so that we ADD BY PORT.
        # For 2: HA group need NOT be created as it will already exist (since
        #        there is already at least on redundancy router). EXTRA PORTS
        #        MUST BE CREATED for each added redundancy router. The id
        #        of extra port should be specified in the interface_info
        #        argument of the add_router_interface call so that we ADD BY
        #        PORT.
        # For 3: if the interface for the user_visible_router was added by ...
        #   a) PORT:   An HA GROUP MUST BE CREATED and and EXTRA PORTS MUST BE
        #              CREATED for each redundancy router. The id of extra port
        #              should be specified in the interface_info argument of
        #              the add_router_interface call so that we ADD BY PORT.
        #   b) SUBNET: There are two cases to consider. If the added interface
        #              of the user_visible_router has ...
        #        b1) 1 SUBNET:   An HA GROUP MUST BE CREATED and and EXTRA
        #                        PORTS MUST BE CREATED for each redundancy
        #                        router. The id of extra port should be
        #                        specified in the interface_info argument of
        #                        the add_router_interface call so we ADD BY
        #                        PORT.
        #        b2) >1 SUBNETS: HA group need NOT be created as it will
        #                        already exist (since the redundancy routers
        #                        should already have extra ports to which the
        #                        (IPv6) subnet is added. Extra ports need
        #                        thus NOT be created. The subnet id should be
        #                        added to the existing extra ports.
        router_id = router['id']
        if ha_settings_db is None:
            ha_settings_db = self._get_ha_settings_by_router_id(context,
                                                                router_id)
        if ha_settings_db is None:
            return
        e_context = context.elevated()
        add_by_subnet = (itfc_info is not None and 'subnet_id' in itfc_info and
                         len(new_port['fixed_ips']) > 1)
        if (add_by_subnet is False or (itfc_info is None and
                                       create_ha_group is True)):
            # generate ha settings and extra port for router (VIP) port
            self._create_ha_group(e_context, router, new_port, ha_settings_db)
        fixed_ips = self._get_fixed_ips_subnets(new_port['fixed_ips'])
        for r_id in (redundancy_router_ids or
                     self._get_redundancy_router_ids(e_context, router_id)):
            if add_by_subnet is True:
                # need to add subnet to redundancy router port
                ports = self._core_plugin.get_ports(
                    e_context,
                    filters={'device_id': [r_id],
                             'network_id': [new_port['network_id']]},
                    fields=['fixed_ips', 'id'])
                redundancy_port = ports[0]
                fixed_ips = redundancy_port['fixed_ips']
                fixed_ip = {'subnet_id': itfc_info['subnet_id']}
                fixed_ips.append(fixed_ip)
                self._core_plugin.update_port(
                    e_context, redundancy_port['id'],
                    {'port': {'fixed_ips': fixed_ips}})
            else:
                redundancy_port = self._create_hidden_port(
                    e_context, new_port['network_id'], '', fixed_ips)
                interface_info = {'port_id': redundancy_port['id']}
                self.add_router_interface(e_context, r_id, interface_info)

    def _update_redundancy_router_interfaces(self, context, router,
                                             port, modified_port_data,
                                             redundancy_router_ids=None,
                                             ha_settings_db=None):
        """To be called when the router interfaces are updated,
        like in the case of change in port admin_state_up status
        """
        router_id = router['id']
        if ha_settings_db is None:
            ha_settings_db = self._get_ha_settings_by_router_id(context,
                                                                router_id)
        if ha_settings_db is None:
            return
        e_context = context.elevated()

        rr_ids = self._get_redundancy_router_ids(e_context, router_id)
        port_info_list = self._core_plugin.get_ports(
            e_context, filters={'device_id': rr_ids,
                                'network_id': [port['network_id']]},
            fields=['device_id', 'id'])
        for port_info in port_info_list:
            self._core_plugin.update_port(e_context, port_info['id'],
                                          modified_port_data)
        self._update_hidden_port(e_context, port['id'], modified_port_data)

    def _create_ha_group(self, context, router, port, ha_settings_db):
        driver = self._get_router_type_driver(context,
                                              router[routertype.TYPE_ATTR])
        if driver is None:
            return
        ha_group_uuid = uuidutils.generate_uuid()
        # use HA group as device instead of the router to hide this port
        group_id = (driver.generate_ha_group_id(context, router, port,
                                                ha_settings_db,
                                                ha_group_uuid) or
                    self._generate_ha_group_id(context, router, port,
                                               ha_settings_db,
                                               ha_group_uuid))
        timers_cfg = driver.get_ha_group_timers_parameters(
            context, router, port, ha_settings_db, ha_group_uuid)
        tracking_cfg = driver.get_ha_group_tracking_parameters(
            context, router, port, ha_settings_db, ha_group_uuid)
        other_cfg = driver.get_other_ha_group_parameters(
            context, router, port, ha_settings_db, ha_group_uuid)
        if driver.ha_interface_ip_address_needed(
                context, router, port, ha_settings_db, ha_group_uuid):
            fixed_ips = self._get_fixed_ips_subnets(port['fixed_ips'])
            extra_port = self._create_hidden_port(
                context, port['network_id'], ha_group_uuid,
                fixed_ips, port['device_owner'])
            extra_port_id = extra_port['id']
        else:
            extra_port_id = None
        subnet_id = (port['fixed_ips'][0]['subnet_id']
                     if port['fixed_ips'] else None)

        with context.session.begin(subtransactions=True):
            r_ha_g = RouterHAGroup(
                id=ha_group_uuid,
                tenant_id=port['tenant_id'],
                ha_type=ha_settings_db.ha_type,
                group_identity=group_id,
                ha_port_id=port['id'],
                extra_port_id=extra_port_id,
                subnet_id=subnet_id,
                user_router_id=router['id'],
                timers_config=timers_cfg,
                tracking_config=tracking_cfg,
                other_config=other_cfg)
            context.session.add(r_ha_g)
        return r_ha_g

    def _get_fixed_ips_subnets(self, fixed_ips):
        subnets = copy.deepcopy(fixed_ips)
        for s in subnets:
            s.pop('ip_address', None)
        return subnets

    def _remove_redundancy_router_interfaces(self, context, router_id,
                                             old_port):
        """To be called in delete_router_interface() BEFORE interface has been
        removed from router in DB.
        """
        ha_settings = self._get_ha_settings_by_router_id(context, router_id)
        if ha_settings is None or old_port is None:
            return
        e_context = context.elevated()
        rr_ids = self._get_redundancy_router_ids(e_context, router_id)
        port_info_list = self._core_plugin.get_ports(
            e_context, filters={'device_id': rr_ids,
                                'network_id': [old_port['network_id']]},
            fields=['device_id', 'fixed_ips', 'id'])
        subnet_id = old_port['fixed_ips'][0]['subnet_id']
        for port_info in port_info_list:
            if port_info['fixed_ips'][0]['subnet_id'] == subnet_id:
                interface_info = {'port_id': port_info['id']}
                self.remove_router_interface(e_context, port_info['device_id'],
                                             interface_info)
        self._delete_ha_group(e_context, old_port['id'])

    def _redundancy_routers_for_floatingip(
            self, context, router_id, redundancy_router_ids=None,
            ha_settings_db=None):
        """To be called in update_floatingip() to get the
            redundant router ids.
        """
        if ha_settings_db is None:
            ha_settings_db = self._get_ha_settings_by_router_id(context,
                                                                router_id)
        if ha_settings_db is None:
            return

        e_context = context.elevated()
        router_ids = []
        for r_id in (redundancy_router_ids or
                     self._get_redundancy_router_ids(e_context, router_id)):
            router_ids.append(r_id)

        return router_ids

    def _update_hidden_port(self, context, ha_port_id, modified_port_data):
        hag = self._get_ha_group_by_ha_port_id(context, ha_port_id)
        if hag is not None:
            self._core_plugin.update_port(context, hag.extra_port_id,
                                          modified_port_data)

    def _delete_ha_group(self, context, ha_port_id):
        hag = self._get_ha_group_by_ha_port_id(context, ha_port_id)
        if hag is not None and hag.extra_port_id is not None:
            self._core_plugin.delete_port(context, hag.extra_port_id,
                                          l3_port_check=False)
            with context.session.begin(subtransactions=True):
                context.session.delete(hag)

    @staticmethod
    @bc.extends([l3_const.ROUTERS])
    def _extend_router_dict_ha(router_res, router_db):
        self = bc.get_plugin(bc.constants.L3)
        if utils.is_extension_supported(self, ha.HA_ALIAS):
            ha_s = router_db.ha_settings
            rr_b = router_db.redundancy_binding
            if rr_b and rr_b.user_router:
                # include static routes from user visible router
                temp = {}
                self._extend_router_dict_extraroute(temp,
                                                    rr_b.user_router)
                if temp['routes']:
                    router_res['routes'].extend(temp['routes'])
            router_res[ha.ENABLED] = False if ha_s is None else True
            if router_res[ha.ENABLED]:
                ha_details = {ha.TYPE: ha_s.ha_type,
                              ha.PRIORITY: ha_s.priority,
                              ha.STATE: ha_s.state,
                              ha.REDUNDANCY_LEVEL: ha_s.redundancy_level,
                              ha.PROBE_CONNECTIVITY: ha_s.probe_connectivity}
                if ha_details[ha.PROBE_CONNECTIVITY]:
                    ha_details.update({ha.PROBE_TARGET: ha_s.probe_target,
                                       ha.PROBE_INTERVAL: ha_s.probe_interval})
                ha_details[ha.REDUNDANCY_ROUTERS] = (
                    [{'id': b.redundancy_router_id, ha.PRIORITY: b.priority,
                      ha.STATE: b.state}
                     for b in router_db.redundancy_bindings])
                router_res[ha.DETAILS] = ha_details
            else:
                # ensure any router details are removed
                router_res.pop(ha.DETAILS, None)

    def _populate_ha_information(self, context, router):
        """To be called when router information, including router interface
        list, (for the l3_cfg_agent) has been collected so it is extended
        with ha information.
        """
        r_r_b = self._get_redundancy_router_bindings(
            context, redundancy_router_id=router['id'])
        if not r_r_b:
            if router[ha.ENABLED]:
                # The router is a user visible router with HA enabled.
                user_router_id = router['id']
                fips = []
            else:
                # The router is a user visible router with HA disabled.
                # Nothing more to do here.
                return
        else:
            # The router is a redundancy router.
            # Need to fetch floatingip configurations from user visible router
            # so they can be added to the redundancy routers.
            user_router_id = r_r_b[0].user_router_id
            fips = self.get_floatingips(context,
                                        {'router_id': [user_router_id]})
        if router['id'] != user_router_id:
            # We add the HA settings from user visible router to
            # its redundancy routers.
            user_router_db = self._get_router(context, user_router_id)
            self._extend_router_dict_ha(router, user_router_db)
        # The interfaces of the user visible router must use the
        # IP configuration of the extra ports in the HA groups.
        hag_dbs = self._get_subnet_id_indexed_ha_groups(context,
                                                        user_router_id)
        e_context = context.elevated()
        if router.get('gw_port'):
            modified_interfaces = []
            interface_port = self._populate_port_ha_information(
                e_context, router['gw_port'], router['id'], hag_dbs,
                user_router_id, modified_interfaces)
            if not interface_port:
                # The router has a gw_port but cannot find the port info yet
                # so mark this router to have incomplete info and bail.
                # The cfg_agent puts this in the updated_routers to ask again.
                router['status'] = cisco_constants.ROUTER_INFO_INCOMPLETE
                return
            if modified_interfaces:
                router['gw_port'] = interface_port
        modified_interfaces = []
        for itfc in router.get(bc.constants.INTERFACE_KEY, []):
            interface_port = self._populate_port_ha_information(
                e_context, itfc, router['id'], hag_dbs, user_router_id,
                modified_interfaces)
            if not interface_port:
                # the router has interfaces but cannot find the port info yet
                # so mark this router to have incomplete info and bail
                # the cfg_agent will put this in the updated_list to ask again
                router['status'] = cisco_constants.ROUTER_INFO_INCOMPLETE
                return
        if modified_interfaces:
            router[bc.constants.INTERFACE_KEY] = modified_interfaces
        if fips:
            router[bc.constants.FLOATINGIP_KEY] = fips

    def _populate_port_ha_information(self, context, port, router_id, hag_dbs,
                                      user_router_id, modified_interfaces):
        subnet_id = port['fixed_ips'][0]['subnet_id']
        try:
            hag_db = hag_dbs[subnet_id]
        except KeyError:
            # Oops, the subnet_id was not found. Probably because the DB
            # insertion of that HA group is still in progress by another
            # process and has not been committed to the DB yet.
            # Let's retry a few times to see if the DB entry turns up.
            LOG.debug('No HA group info for router: %(r_id)s and subnet: '
                      '%(s_id)s was found when populating HA info for port: '
                      '%(p_id)s. Will now make additional lookup attempts.',
                      {'r_id': router_id, 's_id': subnet_id,
                       'p_id': port['id']})
            try:
                hag_db = self._get_ha_group_for_subnet_id(context, router_id,
                                                          subnet_id)
            except exc.NoResultFound:
                hag_db = None
        if hag_db is None:
            LOG.debug('Failed to fetch the HA group info for for router: '
                      '%(r_id)s and subnet: %(s_id)s. Giving up. No HA '
                      'info will be added to the router\'s port: %(p_id)s.',
                      {'r_id': router_id, 's_id': subnet_id,
                       'p_id': port['id']})
            # we leave it to the L3 config agent to handle this
            return
        else:
            LOG.debug('Successfully fetched the HA group info for '
                      'router: %(r_id)s and subnet: %(s_id)s from DB',
                      {'r_id': router_id, 's_id': subnet_id})
            hag_dbs[subnet_id] = hag_db
        if router_id == user_router_id:
            # If the router interface needs no dedicated IP address we just
            # set the HA (VIP) port to the port itself. The config agent
            # driver will know how to handle this "signal".
            p_id = hag_db.extra_port_id or port['id']
            try:
                interface_port = self._core_plugin.get_port(context, p_id)
            except n_exc.PortNotFound:
                LOG.debug('Failed to get port %(p_id)s from DB for user '
                          'visible HA router: %(r_id)s. No HA info will be '
                          'added to that router\'s port', {'p_id': p_id,
                                                           'r_id': router_id})
                return
            LOG.debug('Fetched port %(p_id)s from DB for user visible '
                      'router: %(r_id)s', {'p_id': p_id, 'r_id': router_id})
            self._populate_mtu_and_subnets_for_ports(context, [interface_port])
            modified_interfaces.append(interface_port)
            ha_port = port
        else:
            try:
                ha_port = self._core_plugin.get_port(context,
                                                     hag_db.ha_port_id)
            except n_exc.PortNotFound:
                LOG.debug('Failed to get port %(p_id)s from DB for HA '
                          'redundancy router: %(r_id)s. No HA info will be '
                          'added to that router\'s port',
                          {'p_id': hag_db.ha_port_id, 'r_id': router_id})
                return
            LOG.debug('Fetched port %(p_id)s from DB for HA redundancy '
                      'router: %(r_id)s', {'p_id': hag_db.ha_port_id,
                                           'r_id': router_id})
            self._populate_mtu_and_subnets_for_ports(context, [ha_port])
            interface_port = port
        interface_port[ha.HA_INFO] = {
            ha.TYPE: hag_db.ha_type,
            HA_GROUP: hag_db.group_identity,
            'timers_config': hag_db.timers_config,
            'tracking_config': hag_db.tracking_config,
            'other_config': hag_db.other_config,
            HA_PORT: ha_port}
        return interface_port

    def _create_hidden_port(self, context, network_id, device_id, fixed_ips,
                            port_type=DEVICE_OWNER_ROUTER_INTF):
        """Creates port used specially for HA purposes."""
        port = {'port': {
            'tenant_id': '',  # intentionally not set
            'network_id': network_id,
            'mac_address': ATTR_NOT_SPECIFIED,
            'fixed_ips': fixed_ips,
            'device_id': device_id,
            'device_owner': port_type,
            'admin_state_up': True,
            'name': ''}}
        if utils.is_extension_supported(self._core_plugin, "dns-integration"):
                port['port'].update(dns_name='')
        core_plugin = bc.get_plugin()
        return core_plugin.create_port(context, port)

    def _get_ha_settings_by_router_id(self, context, router_id):
        query = context.session.query(RouterHASetting)
        query = query.filter(RouterHASetting.router_id == router_id)
        try:
            r_ha_s = query.one()
        except (exc.NoResultFound, exc.MultipleResultsFound):
            return
        return r_ha_s

    def _get_ha_group_by_ha_port_id(self, context, port_id):
        query = context.session.query(RouterHAGroup)
        query = query.filter(RouterHAGroup.ha_port_id == port_id)
        try:
            r_ha_g = query.one()
        except (exc.NoResultFound, exc.MultipleResultsFound):
            return
        return r_ha_g

    def _get_subnet_id_indexed_ha_groups(self, context, router_id,
                                         load_virtual_port=False):
        query = context.session.query(RouterHAGroup)
        query = query.filter(RouterHAGroup.user_router_id == router_id)
        if load_virtual_port:
            query = query.options(joinedload('redundancy_router'))
        return {hag['subnet_id']: hag for hag in query}

    @cisco_utils.retry(exc.NoResultFound, LOOKUP_RETRIES, RETRY_INTERVAL, 1)
    def _get_ha_group_for_subnet_id(self, context, router_id, subnet_id):
        query = context.session.query(RouterHAGroup)
        query = query.filter_by(user_router_id=router_id,
                                subnet_id=subnet_id)
        LOG.debug('Trying to fetch HA group info for router: %(r_id)s and '
                  'subnet: %(s_id)s', {'r_id': router_id,
                                       's_id': subnet_id})
        return query.one()

    def _get_redundancy_router_bindings(self, context, router_id=None,
                                        redundancy_router_id=None):
        query = context.session.query(RouterRedundancyBinding)
        if router_id is not None:
            query = query.filter(
                RouterRedundancyBinding.user_router_id == router_id)
        if redundancy_router_id is not None:
            query = query.filter(
                RouterRedundancyBinding.redundancy_router_id ==
                redundancy_router_id)
        query = query.order_by(RouterRedundancyBinding.priority)
        return query.all()

    def _get_redundancy_router_ids(self, context, router_id):
        return [binding.redundancy_router_id for binding in
                self._get_redundancy_router_bindings(context,
                                                     router_id=router_id)]

    def _generate_ha_group_id(self, context, router, port, ha_settings_db,
                              ha_group_uuid):
        #TODO(bob-melander): Generate "guaranteed" unique id
        if ha_settings_db.ha_type == ha.HA_HSRP:
            return random.randint(0, MAX_HSRP_GROUPS)
        elif ha_settings_db.ha_type == ha.HA_VRRP:
            return random.randint(0, MAX_VRRP_GROUPS)
        else:
            # ha_type must be ha_type.GLBP
            return random.randint(0, MAX_GLBP_GROUPS)

    # Overloaded function from l3_db
    def get_router_for_floatingip(self, context, internal_port,
                                  internal_subnet, external_network_id):
        """We need to over-load this function so that we only return the
        user visible router and never its redundancy routers (as they never
        have floatingips associated with them).
        """
        gw_port = orm.aliased(models_v2.Port, name="gw_port")
        routerport_qry = context.session.query(
            RouterPort.router_id, models_v2.IPAllocation.ip_address).join(
            models_v2.Port, models_v2.IPAllocation).filter(
            models_v2.Port.network_id == internal_port['network_id'],
            RouterPort.port_type.in_(bc.constants.ROUTER_INTERFACE_OWNERS),
            models_v2.IPAllocation.subnet_id == internal_subnet['id']
        ).join(gw_port, gw_port.device_id == RouterPort.router_id).filter(
            gw_port.network_id == external_network_id,
            gw_port.device_owner == bc.constants.DEVICE_OWNER_ROUTER_GW
        ).distinct()

        # Ensure that redundancy routers (in a ha group) are not returned,
        # since only the user visible router should have floatingips.
        # This can be done by checking that the id of routers does not
        # appear in the 'redundancy_router_id' column in the
        # 'cisco_router_redundancy_bindings' table.
        routerport_qry = routerport_qry.outerjoin(
            RouterRedundancyBinding,
            RouterRedundancyBinding.redundancy_router_id ==
            RouterPort.router_id)
        routerport_qry = routerport_qry.filter(
            RouterRedundancyBinding.redundancy_router_id == expr.null())

        first_router_id = None
        for router_id, interface_ip in routerport_qry:
            if interface_ip == internal_subnet['gateway_ip']:
                return router_id
            if not first_router_id:
                first_router_id = router_id
        if first_router_id:
            return first_router_id

        raise l3_exceptions.ExternalGatewayForFloatingIPNotFound(
            subnet_id=internal_subnet['id'],
            external_network_id=external_network_id,
            port_id=internal_port['id'])



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class RoutingDriverBase(object):
    """Base class that defines an abstract interface for the Routing Driver.

    This class defines the abstract interface/API for the Routing and
    NAT related operations. Driver class corresponding to a hosting device
    should inherit this base driver and implement its methods.
    RouterInfo object (
    networking_cisco.plugins.cisco.cfg_agent.router_info.RouterInfo)
    is a wrapper around the router dictionary, with attributes for easy access
    to parameters.
    """

    @abc.abstractmethod
    def router_added(self, router_info):
        """A logical router was assigned to the hosting device.

        :param router_info: RouterInfo object for this router
        :returns: None
        """
        pass

    @abc.abstractmethod
    def router_removed(self, router_info):
        """A logical router was de-assigned from the hosting device.

        :param router_info: RouterInfo object for this router
        :returns: None
        """

        pass

    @abc.abstractmethod
    def internal_network_added(self, router_info, port):
        """An internal network was connected to a router.

        :param router_info: RouterInfo object for this router
        :param port: port dictionary for the port where the internal network is
            connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def internal_network_removed(self, router_info, port):
        """An internal network was removed from a router.

        :param router_info: RouterInfo object for this router
        :param port: port dictionary for the port where the internal network
            was connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def external_gateway_added(self, router_info, ext_gw_port):
        """An external network was added to a router.

        :param router_info: RouterInfo object of the router
        :param ext_gw_port: port dictionary for the port where the external
            gateway network is connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def external_gateway_removed(self, router_info, ext_gw_port):
        """An external network was removed from the router.

        :param router_info: RouterInfo object of the router
        :param ext_gw_port: port dictionary for the port where the external
            gateway network was connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def enable_internal_network_NAT(self, router_info, port, ext_gw_port):
        """Enable NAT on an internal network.

        :param router_info: RouterInfo object for this router
        :param port: port dictionary for the port where the internal
            network is connected
        :param ext_gw_port: port dictionary for the port where the external
            gateway network is connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def disable_internal_network_NAT(self, router_info, port, ext_gw_port):
        """Disable NAT on an internal network.

        :param router_info: RouterInfo object for this router
        :param port: port dictionary for the port where the internal network is
            connected
        :param ext_gw_port: port dictionary for the port where the external
            gateway network is connected
        :returns: None
        """
        pass

    @abc.abstractmethod
    def floating_ip_added(self, router_info, ext_gw_port,
                          floating_ip, fixed_ip):
        """A floating IP was added.

        :param router_info: RouterInfo object for this router
        :param ext_gw_port: port dictionary for the port where the external
            gateway network is connected
        :param floating_ip: Floating IP as a string
        :param fixed_ip: Fixed IP of internal internal interface as a string
        :returns: None
        """
        pass

    @abc.abstractmethod
    def floating_ip_removed(self, router_info, ext_gw_port,
                            floating_ip, fixed_ip):
        """A floating IP was removed.

        :param router_info: RouterInfo object for this router
        :param ext_gw_port: port dictionary for the port where the external
                            gateway network is connected
        :param floating_ip: Floating IP as a string
        :param fixed_ip:    Fixed IP of internal internal interface as a string
        :returns: None
        """
        pass

    @abc.abstractmethod
    def routes_updated(self, router_info, action, route):
        """Routes were updated for router.

        :param router_info: RouterInfo object for this router
        :param action: Action on the route , either 'replace' or 'delete'
        :param route: route dictionary with keys 'destination' & 'next_hop'
        :returns: None
        """
        pass

    @abc.abstractmethod
    def cleanup_invalid_cfg(self, hosting_device, routers):
        """Cleanup any invalid configuration in backend.

        :param hosting_device: hosting_device dictionary for backend
        :param routers: list of router dictionaries for routers on backend
        :returns: None
        """
        pass

    @abc.abstractmethod
    def get_configuration(self):
        """Return configuration of hosting_device for driver instance

        :returns: configuration as a text string
        """
        pass



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from mock import patch
import unittest

from networking_cisco.plugins.cisco.cpnr.cpnr_client import CpnrClient


class TestCpnrClient(unittest.TestCase):

    def setUp(self):
        super(TestCpnrClient, self).setUp()
        self.mock_do_request = patch.object(CpnrClient, '_do_request').start()

    def test_buildurl(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        expected_url = ('http://cpnr.com:8080/web-services/rest/'
                        'resource/t?vpnId=vpn1234?viewId=view123&'
                        'zoneOrigin=test.com')
        return_url = mock_client._build_url('t', 'vpn1234', 'view123',
                        'test.com')
        self.assertEqual(expected_url, return_url)

    def test_get_dhcp_server(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)

        mock_client.get_dhcp_server()
        self.mock_do_request.assert_called_once_with('GET',
                    'http://cpnr.com:8080/web-services/rest/'
                    'resource/DHCPServer')

    def test_get_client_classes(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_client_classes()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/ClientClass')

    def test_get_client_class(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_client_class('myclientclass')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'ClientClass/myclientclass')

    def test_get_vpns(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_vpns()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/VPN')

    def test_get_scopes(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_scopes()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/Scope?vpnId=.*')

    def test_get_scope(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_scope('myscope')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/Scope/myscope')

    def test_get_client_entries(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_client_entries()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/ClientEntry')

    def test_get_client_entry(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_client_entry('myclinetentry')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'ClientEntry/myclinetentry')

    def test_get_leases(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_leases('vpn123')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'Lease?vpnId=vpn123')

    def test_get_dns_server(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_dns_server()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/DNSServer')

    def test_get_dns_forwarders(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_dns_forwarders()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/DnsForwarder')

    def test_get_dns_forwarder(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_dns_forwarder('myforwarder')
        self.mock_do_request.assert_called_once_with('GET',
                    'http://cpnr.com:8080/web-services/rest/'
                    'resource/DnsForwarder/myforwarder')

    def test_get_dns_views(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.get_dns_views()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/DnsView')

    def test_get_dns_view(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_dns_view('mydnsview')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'DnsView/mydnsview')

    def test_get_ccm_zones(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_zones()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMZone?viewId=.*')

    def test_get_ccm_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_zone('myzone')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMZone/myzone?viewId=.*')

    def test_get_ccm_reverse_zones(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_reverse_zones()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMReverseZone?viewId=.*')

    def test_get_ccm_reverse_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_reverse_zone('myreversezone')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMReverseZone/myreversezone?viewId=.*')

    def test_get_ccm_hosts(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_hosts()
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMHost?viewId=.*&zoneOrigin=.*')

    def test_get_ccm_host(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.get_ccm_host('myhost')
        self.mock_do_request.assert_called_once_with('GET',
            'http://cpnr.com:8080/web-services/rest/resource/CCMHost'
            '/myhost?viewId=.*&zoneOrigin=.*')

    def test_create_scope(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.create_scope('myscope')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/Scope',
            'myscope')

    def test_create_client_class(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.create_client_class('myclientclass')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/ClientClass',
            'myclientclass')

    def test_create_vpn(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.create_vpn('myvpn')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/VPN',
            'myvpn')

    def test_create_client_entry(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.create_client_entry('mycliententry')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/ClientEntry',
            'mycliententry')

    def test_create_dns_forwarder(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                'admin', 0)
        mock_client.create_dns_forwarder('mydnsforwarder')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/DnsForwarder',
            'mydnsforwarder')

    def test_create_dns_view(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.create_dns_view('mydnsview')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/DnsView',
            'mydnsview')

    def test_create_ccm_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.create_ccm_zone('myccmzone')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/CCMZone',
            'myccmzone')

    def test_create_ccm_reverse_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.create_ccm_reverse_zone('myccmreversezone')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/CCMReverseZone',
            'myccmreversezone')

    def test_create_ccm_host(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.create_ccm_host('myccmhost')
        self.mock_do_request.assert_called_once_with('POST',
            'http://cpnr.com:8080/web-services/rest/resource/CCMHost',
            'myccmhost')

    def test_update_dhcp_server(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_dhcp_server('updatedhcpserver')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/DHCPServer',
            'updatedhcpserver')

    def test_update_client_class(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_client_class('updateclientclass',
                                        'newclientclass')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'ClientClass/updateclientclass',
            'newclientclass')

    def test_update_vpn(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_vpn('updatevpn', 'newvpn')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/VPN/updatevpn',
            'newvpn')

    def test_update_scope(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_scope('updatescope', 'newscope')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'Scope/updatescope',
            'newscope')

    def test_update_client_entry(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_client_entry('updatecliententry',
                                        'newcliententry')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource'
            '/ClientEntry/updatecliententry',
            'newcliententry')

    def test_update_dns_server(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_dns_server('updatednsserver')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/DNSServer',
            'updatednsserver')

    def test_update_dns_forwarder(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_dns_forwarder('updatednsforwarder',
                                         'newforwarder')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'DnsForwarder/updatednsforwarder',
            'newforwarder')

    def test_update_dns_view(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_dns_view('updatednsview', 'newdnsview')
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource'
            '/DnsView/updatednsview',
            'newdnsview')

    def test_update_ccm_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_ccm_zone('updateccmzone', 'newzone',
                                    None)
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMZone/updateccmzone',
            'newzone')

    def test_update_ccm_reverse_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_ccm_reverse_zone(
                                        'updateccmreversezone',
                                        'newreversezone',
                                        None)
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMReverseZone/updateccmreversezone',
            'newreversezone')

    def test_update_ccm_host(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.update_ccm_host('updateccmhost', 'newccmhost',
                                    None)
        self.mock_do_request.assert_called_once_with('PUT',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMHost/updateccmhost',
            'newccmhost')

    def test_delete_client_class(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_client_class('deleteclientclass')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'ClientClass/deleteclientclass')

    def test_delete_vpn(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_vpn('deletevpn')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/VPN/deletevpn')

    def test_delete_scope(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_scope('deletescope')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'Scope/deletescope')

    def test_delete_client_entry(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_client_entry('deletecliententry')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'ClientEntry/deletecliententry')

    def test_delete_dns_forwarder(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_dns_forwarder('deletednsforwarder')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'DnsForwarder/deletednsforwarder')

    def test_delete_dns_view(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_dns_view('deletednsview')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'DnsView/deletednsview')

    def test_delete_ccm_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_ccm_zone('deleteccmzone')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMZone/deleteccmzone')

    def test_delete_ccm_reverse_zone(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_ccm_reverse_zone('delccmreversezone')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMReverseZone/delccmreversezone')

    def test_delete_ccm_host(self):
        mock_client = CpnrClient('http', 'cpnr.com', '8080', 'admin',
                                 'admin', 0)
        mock_client.delete_ccm_host('deleteccmhost')
        self.mock_do_request.assert_called_once_with('DELETE',
            'http://cpnr.com:8080/web-services/rest/resource/'
            'CCMHost/deleteccmhost')



# Copyright (c) 2017 Cisco Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from networking_cisco import backwards_compatibility as bc
from networking_cisco.ml2_drivers.nexus import (
    nexus_helpers as nexus_help)

LOG = log.getLogger(__name__)


class NexusMDTrunkHandler(object):
    """Cisco Nexus Mechanism Driver Trunk Handler.

    This class contains methods called by the cisco_nexus MD for
    processing trunk subports.
    """
    def is_trunk_parentport(self, port):
        return 'trunk_details' in port

    def is_trunk_subport(self, port):
        return port['device_owner'] == bc.trunk_consts.TRUNK_SUBPORT_OWNER

    def is_trunk_subport_baremetal(self, port):
        context = bc.get_context()
        el_context = context.elevated()

        subport_obj = bc.trunk_objects.SubPort.get_object(
            el_context, port_id=port['id'])
        if subport_obj:
            trunk_obj = bc.trunk_objects.Trunk.get_object(
                el_context, id=subport_obj.trunk_id)
            trunk_port = bc.get_plugin().get_port(
                el_context, trunk_obj.port_id)
            return nexus_help.is_baremetal(trunk_port)
        else:
            return False

    def update_subports(self, port):
        """Set port attributes for trunk subports.

        For baremetal deployments only, set the neutron port attributes
        during the bind_port event.
        """
        trunk_details = port.get('trunk_details')
        subports = trunk_details['sub_ports']

        host_id = port.get(bc.dns.DNSNAME)
        context = bc.get_context()
        el_context = context.elevated()

        for subport in subports:
            bc.get_plugin().update_port(el_context, subport['port_id'],
                {'port':
                 {bc.portbindings.HOST_ID: host_id,
                  'device_owner': bc.trunk_consts.TRUNK_SUBPORT_OWNER}})

        # Set trunk to ACTIVE status.
        trunk_obj = bc.trunk_objects.Trunk.get_object(
            el_context, id=trunk_details['trunk_id'])
        trunk_obj.update(status=bc.trunk_consts.ACTIVE_STATUS)



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from oslo_config import cfg
from oslo_utils import importutils

from neutron.api.rpc.agentnotifiers import dhcp_rpc_agent_api
from neutron.common import topics
from neutron.db import agentschedulers_db
from neutron.plugins.ml2 import rpc
from neutron.tests.unit.db import test_agentschedulers_db

from networking_cisco import backwards_compatibility as bc
from networking_cisco.plugins.cisco.common import cisco_constants as c_const
from networking_cisco.plugins.cisco.extensions import (
    ciscohostingdevicemanager as ciscodevmgr)
from networking_cisco.tests.unit.cisco.device_manager import (
    device_manager_test_support)
from networking_cisco.tests.unit.cisco.device_manager import (
    test_db_device_manager)
from networking_cisco.tests.unit.cisco.l3 import (
    test_l3_routertype_aware_schedulers)
from networking_cisco.tests.unit.cisco.l3 import l3_router_test_support
from networking_cisco.tests.unit.cisco.l3 import test_db_routertype


L3_HOSTA = test_agentschedulers_db.L3_HOSTA
L3_HOSTB = test_agentschedulers_db.L3_HOSTB

CORE_PLUGIN_KLASS = ('networking_cisco.tests.unit.cisco.l3'
                     '.test_agent_scheduler.TestAgentSchedCorePlugin')
L3_PLUGIN_KLASS = test_l3_routertype_aware_schedulers.L3_PLUGIN_KLASS


# A core plugin supporting Cisco device manager functionality
class TestAgentSchedCorePlugin(device_manager_test_support.TestCorePlugin,
                               agentschedulers_db.DhcpAgentSchedulerDbMixin):

    supported_extension_aliases = ["external-net", "agent",
                                   bc.constants.DHCP_AGENT_SCHEDULER_EXT_ALIAS,
                                   ciscodevmgr.HOSTING_DEVICE_MANAGER_ALIAS]

    def __init__(self):
        self.network_scheduler = importutils.import_object(
            cfg.CONF.network_scheduler_driver)
        super(TestAgentSchedCorePlugin, self).__init__()
        self._start_rpc_notifiers()

    def _start_rpc_notifiers(self):
        """Initialize RPC notifiers for agents."""
        self.notifier = rpc.AgentNotifierApi(topics.AGENT)
        self.agent_notifiers[bc.constants.AGENT_TYPE_DHCP] = (
            dhcp_rpc_agent_api.DhcpAgentNotifyAPI())


class L3RouterApplianceL3AgentSchedulerTestCase(
    test_agentschedulers_db.OvsAgentSchedulerTestCase,
    test_db_routertype.RoutertypeTestCaseMixin,
    test_db_device_manager.DeviceManagerTestCaseMixin,
    l3_router_test_support.L3RouterTestSupportMixin,
        device_manager_test_support.DeviceManagerTestSupportMixin):

    resource_prefix_map = (test_db_device_manager.TestDeviceManagerDBPlugin
                           .resource_prefix_map)

    def setup_coreplugin(self, core_plugin=None, load_plugins=True):
        # NOTE(bobmel): we override this function to make sure our core plugin,
        # which includes device manager functionality, is used for the tests
        cp_str = CORE_PLUGIN_KLASS if core_plugin == "ml2" else core_plugin
        if bc.NEUTRON_VERSION.version[0] <= \
                bc.NEUTRON_NEWTON_VERSION.version[0]:
            super(L3RouterApplianceL3AgentSchedulerTestCase,
                  self).setup_coreplugin(cp_str)
        else:
            super(L3RouterApplianceL3AgentSchedulerTestCase,
                  self).setup_coreplugin(cp_str, load_plugins)

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        self.plugin_str = (CORE_PLUGIN_KLASS if core_plugin is None
                           else core_plugin)
        self.l3_plugin = (L3_PLUGIN_KLASS if l3_plugin is None
                          else l3_plugin)

        self._define_keystone_authtoken()
        cfg.CONF.set_override('api_extensions_path',
                              l3_router_test_support.extensions_path)
        cfg.CONF.set_override('default_router_type',
                              c_const.NAMESPACE_ROUTER_TYPE, group='routing')

        super(L3RouterApplianceL3AgentSchedulerTestCase, self).setUp()

        self._mock_l3_admin_tenant()
        templates = self._test_create_hosting_device_templates()
        self._test_create_routertypes(templates.values())
        # mock the periodic router backlog processing in the tests
        self._mock_backlog_processing(self.l3plugin)

    def tearDown(self):
        self._test_remove_routertypes()
        self._test_remove_hosting_device_templates()
        super(L3RouterApplianceL3AgentSchedulerTestCase, self).tearDown()

    @unittest.skip("DVR not supported")
    def test_dvr_router_scheduling_to_only_dvr_snat_agent(self):
        pass

    @unittest.skip("DVR not supported")
    def test_dvr_router_csnat_rescheduling(self):
        pass

    @unittest.skip("DVR not supported")
    def test_dvr_router_manual_rescheduling(self):
        pass

    @unittest.skip("DVR not supported")
    def test_sync_dvr_router(self):
        pass


class L3RouterApplianceL3AgentNotifierTestCase(
    test_agentschedulers_db.OvsL3AgentNotifierTestCase,
    test_db_routertype.RoutertypeTestCaseMixin,
    test_db_device_manager.DeviceManagerTestCaseMixin,
    l3_router_test_support.L3RouterTestSupportMixin,
        device_manager_test_support.DeviceManagerTestSupportMixin):

    resource_prefix_map = (test_db_device_manager.TestDeviceManagerDBPlugin
                           .resource_prefix_map)

    def setup_coreplugin(self, core_plugin=None, load_plugins=True):
        # NOTE(bobmel): we override this function to make sure our core plugin,
        # which includes device manager functionality, is used for the tests
        cp_str = CORE_PLUGIN_KLASS if core_plugin == "ml2" else core_plugin
        if bc.NEUTRON_VERSION.version[0] <= \
                bc.NEUTRON_NEWTON_VERSION.version[0]:
            super(L3RouterApplianceL3AgentNotifierTestCase,
                  self).setup_coreplugin(cp_str)
        else:
            super(L3RouterApplianceL3AgentNotifierTestCase,
                  self).setup_coreplugin(cp_str, load_plugins)

    def setUp(self, core_plugin=None, l3_plugin=None, dm_plugin=None,
              ext_mgr=None):
        self.plugin_str = (CORE_PLUGIN_KLASS if core_plugin is None
                           else core_plugin)
        self.l3_plugin = (L3_PLUGIN_KLASS if l3_plugin is None
                          else l3_plugin)

        self._define_keystone_authtoken()
        cfg.CONF.set_override('api_extensions_path',
                              l3_router_test_support.extensions_path)
        cfg.CONF.set_override('default_router_type',
                              c_const.NAMESPACE_ROUTER_TYPE, group='routing')

        super(L3RouterApplianceL3AgentNotifierTestCase, self).setUp()

        self._mock_l3_admin_tenant()
        templates = self._test_create_hosting_device_templates()
        self._test_create_routertypes(templates.values())

    def tearDown(self):
        self._test_remove_routertypes()
        self._test_remove_hosting_device_templates()
        super(L3RouterApplianceL3AgentNotifierTestCase, self).tearDown()



# Copyright 2016 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
netns - context manager for network namespaces
"""

import ctypes
import os
import platform
import resource
import subprocess

from oslo_log import log as logging

LOG = logging.getLogger(__name__)

# OSX uses dylib extension for shared libraries. This is required for Unit
# Tests to run on OSX.
if platform.uname()[0] == 'Darwin':
    _libc = ctypes.CDLL('libc.dylib')
else:
    _libc = ctypes.CDLL('libc.so.6')

NETNS_DIR = "/var/run/netns/"


class Namespace(object):
    def __init__(self, name):
        self.parent_fd = open("/proc/self/ns/net")
        self.parent_fileno = self.parent_fd.fileno()
        self.target_fd = open(NETNS_DIR + str(name))
        self.target_fileno = self.target_fd.fileno()

    def __enter__(self):
        _libc.setns(self.target_fileno, 0)

    def __exit__(self, type, value, tb):
        _libc.setns(self.parent_fileno, 0)
        try:
            self.target_fd.close()
        except Exception:
            LOG.warning("Failed to close target_fd: %s", self.target_fd)
            pass
        self.parent_fd.close()


def nslist():
    return os.listdir(NETNS_DIR) if os.path.exists(NETNS_DIR) else []


def iflist(ignore=set()):
    interfaces = []
    for line in subprocess.check_output(['ip', 'addr', 'show']).splitlines():
        if not line.strip().startswith(b'inet '):
            continue
        words = line.split()
        name = words[-1]
        if name in ignore:
            continue
        addr, _, mask = words[1].partition(b'/')
        interfaces.append((name, addr, mask))
    return interfaces


def increase_ulimit(ulimit):
    resource.setrlimit(resource.RLIMIT_NOFILE, (ulimit, ulimit))



# Copyright 2014 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_log import log as logging
import webob.exc

from neutron.api import extensions
from neutron.api.v2 import resource
from neutron.common import rpc as n_rpc
from neutron import policy
from neutron import wsgi
from neutron_lib import exceptions

from networking_cisco._i18n import _
from networking_cisco import backwards_compatibility as bc
from networking_cisco.backwards_compatibility import agent_exceptions
from networking_cisco.backwards_compatibility import cb_faults
from networking_cisco.plugins.cisco.common import cisco_constants
from networking_cisco.plugins.cisco.extensions import ciscohostingdevicemanager

PATH_PREFIX = "/dev_mgr"

LOG = logging.getLogger(__name__)


class InvalidCfgAgent(agent_exceptions.AgentNotFound):
    message = _("Agent %(agent_id)s is not a Cisco cfg agent or has been "
                "disabled")


class HostingDeviceAssignedToCfgAgent(exceptions.Conflict):
    message = _("The hosting device %(hosting_device_id)s is already assigned "
                "to Cisco cfg agent %(agent_id)s.")


class HostingDeviceSchedulingFailed(exceptions.Conflict):
    message = _("Failed to assign hosting device %(hosting_device_id)s to "
                "Cisco cfg agent %(agent_id)s.")


class HostingDeviceNotAssignedToCfgAgent(exceptions.NotFound):
    message = _("The hosting device %(hosting_device_id)s is currently not "
                "assigned to Cisco cfg agent %(agent_id)s.")


CFG_AGENT_SCHEDULER_ALIAS = 'cisco-cfg-agent-scheduler'
CFG_AGENT_HOSTING_DEVICE = 'cfg-agent-hosting-device'
CFG_AGENT_HOSTING_DEVICES = CFG_AGENT_HOSTING_DEVICE + 's'
HOSTING_DEVICE_CFG_AGENT = 'hosting-device-cfg-agent'
HOSTING_DEVICE_CFG_AGENTS = HOSTING_DEVICE_CFG_AGENT + 's'


class HostingDeviceSchedulerController(wsgi.Controller):
    def get_plugin(self):
        plugin = bc.get_plugin(cisco_constants.DEVICE_MANAGER)
        if not plugin:
            LOG.error('No Device manager service plugin registered to '
                      'handle hosting device scheduling')
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context, "get_%s" % CFG_AGENT_HOSTING_DEVICES,
                       {})
        return plugin.list_hosting_devices_handled_by_cfg_agent(
            request.context, kwargs['agent_id'])

    def create(self, request, body, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context, "create_%s" % CFG_AGENT_HOSTING_DEVICE,
                       {})
        cfg_agent_id = kwargs['agent_id']
        hosting_device_id = body['hosting_device_id']
        result = plugin.assign_hosting_device_to_cfg_agent(
            request.context, cfg_agent_id, hosting_device_id)
        notify(request.context, 'agent.hosting_device.add', hosting_device_id,
               cfg_agent_id)
        return result

    def delete(self, request, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context, "delete_%s" % CFG_AGENT_HOSTING_DEVICE,
                       {})
        cfg_agent_id = kwargs['agent_id']
        hosting_device_id = kwargs['id']
        result = plugin.unassign_hosting_device_from_cfg_agent(
            request.context, cfg_agent_id, hosting_device_id)
        notify(request.context, 'agent.hosting_device.remove',
               hosting_device_id, cfg_agent_id)
        return result


class CfgAgentsHandlingHostingDeviceController(wsgi.Controller):
    def get_plugin(self):
        plugin = bc.get_plugin(cisco_constants.DEVICE_MANAGER)
        if not plugin:
            LOG.error('No device manager service plugin registered to '
                      'handle hosting device scheduling')
            msg = _('The resource could not be found.')
            raise webob.exc.HTTPNotFound(msg)
        return plugin

    def index(self, request, **kwargs):
        plugin = self.get_plugin()
        policy.enforce(request.context, "get_%s" % HOSTING_DEVICE_CFG_AGENTS,
            {})
        return plugin.list_cfg_agents_handling_hosting_device(
            request.context, kwargs['hosting_device_id'])


class Ciscocfgagentscheduler(bc.extensions.ExtensionDescriptor):
    """Extension class supporting configuration agent scheduler."""
    @classmethod
    def get_name(cls):
        return "Cisco Configuration Agent Scheduler"

    @classmethod
    def get_alias(cls):
        return CFG_AGENT_SCHEDULER_ALIAS

    @classmethod
    def get_description(cls):
        return "Schedule hosting devices among Cisco configuration agents"

    @classmethod
    def get_namespace(cls):
        return ("http://docs.openstack.org/ext/" +
                CFG_AGENT_SCHEDULER_ALIAS + "/api/v1.0")

    @classmethod
    def get_updated(cls):
        return "2014-03-31T10:00:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Ext Resources."""
        exts = []
        parent = dict(member_name="agent",
                      collection_name="agents")
        controller = resource.Resource(HostingDeviceSchedulerController(),
                                       cb_faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(CFG_AGENT_HOSTING_DEVICES,
                                                 controller, parent))
        parent = dict(member_name=ciscohostingdevicemanager.DEVICE,
                      collection_name=ciscohostingdevicemanager.DEVICES)
        controller = resource.Resource(
            CfgAgentsHandlingHostingDeviceController(), cb_faults.FAULT_MAP)
        exts.append(extensions.ResourceExtension(HOSTING_DEVICE_CFG_AGENTS,
                                                 controller, parent,
                                                 PATH_PREFIX))
        return exts

    def get_extended_resources(self, version):
        return {}


class CfgAgentSchedulerPluginBase(object):
    """REST API to operate the cfg agent scheduler.

    All of method must be in an admin context.
    """
    @abc.abstractmethod
    def assign_hosting_device_to_cfg_agent(self, context, id,
                                           hosting_device_id):
        pass

    @abc.abstractmethod
    def unassign_hosting_device_from_cfg_agent(self, context, id,
                                               hosting_device_id):
        pass

    @abc.abstractmethod
    def list_hosting_devices_handled_by_cfg_agent(self, context, id):
        pass

    @abc.abstractmethod
    def list_cfg_agents_handling_hosting_device(self, context,
                                                hosting_device_id):
        pass


def notify(context, action, hosting_device_id, cfg_agent_id):
    info = {'id': cfg_agent_id, 'hosting_device_id': hosting_device_id}
    notifier = n_rpc.get_notifier('hosting_device')
    notifier.info(context, action, {'cfg_agent': info})



# Copyright 2015-2016 Cisco Systems, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# VIF_DETAILS attribute for port profile name used by the Nova VIF driver.
VIF_DETAILS_PROFILEID = 'profileid'

# Supported network interface cards
PCI_INFO_CISCO_VIC_1240 = "1137:0071"
PCI_INFO_INTEL_82599 = "8086:10c9"

VLAN_PATH = "fabric/lan"
VLAN_COMPRESSION_TYPE = "included"
DESCR = "Created by OpenStack UCSM Mech Driver"
PORT_PROFILESETDN = "fabric/lan/profiles"

# High Performance mode to be set within Port profiles for "Direct" mode.
HIGH_PERF = "high-perf-reqd"
NONE = "none"

# Max ports that this port profile can be applied to
MAX_PORTS = 64

VIF_TYPE_802_QBH = '802.1qbh'

VNIC_PATH_PREFIX = "/vnic-"
VLAN_PATH_PREFIX = "/if-"
VLAN_PROFILE_PATH_PREFIX = "/net-"
VLAN_PROFILE_NAME_PREFIX = "OS-"
PORT_PROFILE_NAME_PREFIX = "OS-PP-"
CLIENT_PROFILE_NAME_PREFIX = "OS-CL-"
CLIENT_PROFILE_PATH_PREFIX = "/cl-"
SERVICE_PROFILE_PATH_PREFIX = "org-root/ls-"

ETH0 = "eth0"
ETH1 = "eth1"
ETH_PREFIX = "/ether-"
DUPLICATE_EXCEPTION = "object already exists"

SP_TEMPLATE_PARENT_DN = "org-root/ls-"
SP_TEMPLATE_PREFIX = "/ls-"
VNIC_TEMPLATE_PARENT_DN = "org-root"
VNIC_TEMPLATE_PREFIX = "/lan-conn-templ-"
TRANSPORT = "ethernet"

DEFAULT_PP_DELETE_TIME = (10 * 60)
MAX_PP_DELETE_RETRY_COUNT = 3

SSL_WARNING = ("SSL certificate verification has been disabled for the UCSM "
               "driver. The connection to UCS Manager(s) is now insecure.")
